#ifndef INCLUDED_BODY
#define INCLUDED_BODY

#include "global.h"
template<class T> class List;
class View3D;

//------------------------------------------------------------------------------------------------
//	Classes for a body
//------------------------------------------------------------------------------------------------

class Object;
typedef void (Object::*pfnObj)(struct Tcl_Obj *objv[]);
class Body;

class Object {
	int id, drawId;
	GLfloat red, green, blue;
	GLfloat mat_specular[3];
	GLfloat mat_shininess[1];

	int selected, wireframe;

public:
	Object();
	~Object() {}

	void SetId(int ident) { drawId = id = ident; }
	void SetDrawId(int ident) { drawId = ident; }
	int  GetId() const { return id; }

	void Wireframe(struct Tcl_Obj *objv[]);
	void Shade(struct Tcl_Obj *objv[]);
	void GetShaded(struct Tcl_Obj *objv[]);
	void IsShaded(struct Tcl_Obj *objv[]);
	void Select(struct Tcl_Obj *objv[]);
	void Unselect(struct Tcl_Obj *objv[]);
	void IsSelected(struct Tcl_Obj *objv[]);
	void GetColor(struct Tcl_Obj *objv[]);
	void SetColor(struct Tcl_Obj *objv[]);

	void DoOperation(pfnObj op, struct Tcl_Obj *objv[]) { (this->*op)(objv); }
	void Draw() const;
};

//------------------------------------------------------------------------------------------------

class Leg {
	class Object thigh, leg, foot, toe;

	friend class Body;
	friend int Body_ObjCmd(ClientData clientData, Tcl_Interp *interp,
							int objc, Tcl_Obj *CONST objv[]);

public:
	void DoOperation(pfnObj op, struct Tcl_Obj *objv[]);
	void Draw(const class Body *body) const;
};

class Arm {
	class Object arm, forearm, hand;

	friend class Body;
	friend int Body_ObjCmd(ClientData clientData, Tcl_Interp *interp,
							int objc, Tcl_Obj *CONST objv[]);

public:
	void DoOperation(pfnObj op, struct Tcl_Obj *objv[]);
	void Draw(const class Body *body) const;
};

//------------------------------------------------------------------------------------------------

// related with jointName in body.cc
enum jointEnum {
	bTorso, bPelvis, bNeck, bHead, bHeadHair,
	lHip, lKnee, lAnkle, lFoot,
	lShoulder, lElbow, lWrist,
	rHip, rKnee, rAnkle, rFoot,
	rShoulder, rElbow, rWrist,
	cXZ, cY, cRX, cRZ, cRY,
	// interface for pin to floor
	pin };

enum trackEnum {
	pelvisX, pelvisY, pelvisZ,							/*  0 */
	torsoX, torsoY, torsoZ,								/*  3 */
	headX, headY, headZ,								/*  6 */
	// left
	lShoulderX, lShoulderY, lShoulderZ, lShoulderTY,	/*  9 */
	lElbowX, lElbowY,									/* 13 */
	lWristX, lWristY, lWristZ,							/* 15 */
	lHipX, lHipY, lHipZ,								/* 18 */
	lKneeX,												/* 21 */
	lAnkleX, lAnkleY, lAnkleZ,							/* 22 */
	lFootX,												/* 25 */
	// right
	rShoulderX, rShoulderY, rShoulderZ, rShoulderTY,	/* 26 */
	rElbowX, rElbowY,									/* 30 */
	rWristX, rWristY, rWristZ,							/* 32 */
	rHipX, rHipY, rHipZ,								/* 35 */
	rKneeX,												/* 38 */
	rAnkleX, rAnkleY, rAnkleZ,							/* 39 */
	rFootX,												/* 42 */
	// position
	cogX, cogY, cogZ,									/* 43 */
	cogRX, cogRY, cogRZ };								/* 46 */

enum { left = lShoulderX, right = rShoulderX };
enum {
	shoulderX, shoulderY, shoulderZ, shoulderTY,
	elbowX, elbowY,
	wristX, wristY, wristZ,
	hipX, hipY, hipZ,
	kneeX,
	ankleX, ankleY, ankleZ,
	footX };


class Body {
	// interface
	int number;
	char *name;
	int sex, ghost;
	double joints[NB_DOF];
	int constraint;

	// object (limbs)
	class Object torso, pelvis, neck, headFace, headHair;
	class Arm Larm, Rarm;
	class Leg Lleg, Rleg;

	friend class View3D;
	friend class Leg;
	friend class Arm;
	friend class Interpolator;
	friend class Routine;
	friend int Body_ObjCmd(ClientData clientData, Tcl_Interp *interp,
							int objc, Tcl_Obj *CONST objv[]);

public:
	// what to update when the body changes
	class List<Togl> *views;

	Body(int nb, char *n);
	~Body();

	int Number() const { return number; }
	void DoOperation(pfnObj op, struct Tcl_Obj *objv[]);
	void Draw(int draw_mode, int jointsOnly, int withRX, int withRY, int withRZ) const;
	inline void Redisplay() const;
	double Position(int what, double *x, double *y, double *z,
					int jointsOnly, int withRX, int withRY, int withRZ) const;

	int Set(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
	int Get(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) const;
	int Limits(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) const;
	int Position(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) const;
};


#endif
