namespace eval main {
global status

	variable data
	set data(v:currentDir) $status(dir:routine)
	set data(v:filename) ""
	set data(v:firstSave) 0
}

#=================================================================================================

# load the icons for the toolbar
image create pixmap imgNew -file pixmap/new.xpm
image create pixmap imgOpen -file pixmap/open.xpm
image create pixmap imgSave -file pixmap/save.xpm

image create pixmap imgLibrary -file pixmap/library.xpm
image create pixmap imgRoutine -file pixmap/routine.xpm
image create pixmap imgStage -file pixmap/stage.xpm
image create pixmap imgMovie -file pixmap/movie.xpm

image create pixmap imgProperty -file pixmap/property.xpm

catch {image create photo imgAbout -file $status(dir:logo.gif)}


#-------------------
# define the toolbar
toolbar::new .main
pack .main -side left -pady 5 -padx 5

# add the controls
toolbar::add:button .main new -bd 1 -image imgNew -command "main::open new" -tip "New routine"
toolbar::add:button .main open -bd 1 -image imgOpen -command "main::open load" -tip "Load routine"
toolbar::add:button .main save -bd 1 -image imgSave -command "main::save current" \
		-tip "Save routine"

toolbar::add:separator .main 5

toolbar::add:check .main library -bd 1 -image imgLibrary -tip "Toggle library view" \
		-oncommand "wm deiconify .library" -offcommand "wm withdraw .library" \
		-variable main::data(v:check:library)
toolbar::add:check .main routine -bd 1 -image imgRoutine -tip "Toggle routine editor view" \
		-oncommand "wm deiconify .routine" -offcommand "wm withdraw .routine" \
		-variable main::data(v:check:routine)
toolbar::add:check .main stage -bd 1 -image imgStage -tip "Toggle stage view" \
		-oncommand "wm deiconify .stage" -offcommand "wm withdraw .stage" \
		-variable main::data(v:check:stage)
toolbar::add:check .main movie -bd 1 -image imgMovie -tip "Toggle movie view" \
		-oncommand "wm deiconify .movie" -offcommand "wm withdraw .movie" \
		-variable main::data(v:check:movie)

toolbar::add:separator .main 5

toolbar::add:button .main property -bd 1 -image imgProperty \
		-command "routine::property .routine" -tip "Edit properties of the routine"

#----------------
# define the menu
set m [menu .mainmenu -font menubutton -bd 1 -activeborderwidth 1]
  $m add cascade -label "File" -menu $m.file
	menu $m.file -font menubutton -tearoff 0 -bd 1 -activeborderwidth 1
	$m.file add command -label "New" -command "main::open new" -accelerator "Ctrl-N"
	$m.file add command -label "New couple" -command "main::new:couple"
	$m.file add separator
	$m.file add command -label "Open ..." -command "main::open load" -accelerator "Ctrl-O"
	$m.file add command -label "Save" -command "main::save current" -accelerator "Ctrl-S"
	$m.file add command -label "Save as ..." -command "main::save new"
	$m.file add command -label "Close" -command "main::close" -accelerator "Ctrl-W"
	$m.file add separator
	$m.file add command -label "Quit" -command exit -accelerator "Ctrl-Q"

  $m add cascade -label "Windows" -menu $m.win
	menu $m.win -font menubutton -tearoff 0 -bd 1 -activeborderwidth 1
	$m.win add checkbutton -label " Library" -variable main::data(v:check:library)
	$m.win add checkbutton -label " Editor" -variable main::data(v:check:routine)
	$m.win add checkbutton -label " Stage" -variable main::data(v:check:stage)
	$m.win add checkbutton -label " Movie" -variable main::data(v:check:movie)
	$m.win add separator
	$m.win add command -label "Properties ..." -command "routine::property .routine"

  $m add cascade -label "?" -menu $m.help
	menu $m.help -font menubutton -tearoff 0 -bd 1 -activeborderwidth 1
	$m.help add command -label "About ..." -command "main::about"
	$m.help add command -label "Online manual" -command "main::manual"


. configure -menu $m
wm title . "Gepetto"
bind all <Control-n> "main::open new"
bind all <Control-o> "main::open load"
bind all <Control-s> "main::save current"
bind all <Control-w> "main::close"
bind all <Control-q> exit

# initial configuration
toolbar::configure .main save -state disabled
toolbar::configure .main routine -state disabled
toolbar::configure .main stage -state disabled
toolbar::configure .main movie -state disabled
toolbar::configure .main property -state disabled
.mainmenu.file entryconfigure 4 -state disabled
.mainmenu.file entryconfigure 5 -state disabled
.mainmenu.file entryconfigure 6 -state disabled
.mainmenu.win entryconfigure 1 -state disabled
.mainmenu.win entryconfigure 2 -state disabled
.mainmenu.win entryconfigure 3 -state disabled
.mainmenu.win entryconfigure 5 -state disabled


#=================================================================================================

proc main::open {type {filename {}}} {
global status
variable data

	# check if current was modified
	if {[winfo exists .routine] && [routine::wasModified .routine]} {
		set dosave [tk_messageBox -icon "question" -type yesno -default yes -title "Error" \
						-message "The routine was modified.\nSave it before ?"]
		if {$dosave == "yes"} {
			save current
		}
	}

	# type of openning : new routine or load file
	if {$type == "load" && $filename == ""} {
		set filename [tk_getOpenFile -initialdir "$data(v:currentDir)" \
							-title "Choose routine to open"]
		if {$filename == ""} {
			return
		}
	}
	movie::close .movie
	destroy .movie
	stage::close .stage
	destroy .stage
	routine::closeWindow .routine
	destroy .routine

	# dialog for waiting
	destroy .wait
	toplevel .wait -class Dialog
	wm iconname .wait Dialog
	wm title .wait "Information"
	wm protocol .wait WM_DELETE_WINDOW { }
	wm transient .wait .
	frame .wait.top
	pack .wait.top -side top -fill both -expand 1
	.wait.top configure -relief raised -bd 1
	option add *Dialog.msg.wrapLength 3i widgetDefault
	if {$type == "new"} {
		label .wait.msg -justify left -text "Please wait, creating new routine ..."
	} else {
		label .wait.msg -justify left -text "Please wait, loading routine ..."
	}
	catch {.wait.msg configure -font -Adobe-Times-Medium-R-Normal--*-180-*-*-*-*-*-*}
	pack .wait.msg -in .wait.top -side right -expand 1 -fill both -padx 3m -pady 3m
	label .wait.bitmap -image dlg_info
	pack .wait.bitmap -in .wait.top -side left -padx 3m -pady 3m
	wm withdraw .wait
	update idletasks
	set x [expr [winfo screenwidth .]/2 - [winfo reqwidth .wait]/2 \
			- [winfo vrootx [winfo parent .wait]]]
	set y [expr [winfo screenheight .]/2 - [winfo reqheight .wait]/2 \
			- [winfo vrooty [winfo parent .wait]]]
	wm deiconify .wait
	wm geometry .wait +$x+$y
	update

	# new routine editor
	routine::new .routine $status(pref:routine:pos) $status(pref:routine:size)
	if {$type == "new"} {
		set data(v:currentDir) $status(dir:routine)
		set data(v:filename) ""
		routine::load .routine ""
	} else {
		set data(v:currentDir) [file dirname $filename]
		set data(v:filename) $filename
		routine::load .routine $filename
	}
	if {[toolbar::isOff .main routine]} {
		wm withdraw .routine
	}
	set data(v:firstSave) 1
	# new stage
	stage::new .stage .routine $status(pref:stage:pos) $status(pref:stage:size)
	stage::add .stage [routine::getBodies .routine]
	if {[toolbar::isOff .main stage]} {
		wm withdraw .stage
	}
	# new movie
	movie::new .movie .routine $status(pref:movie:pos) $status(pref:movie:size)
	if {[toolbar::isOff .main movie]} {
		wm withdraw .movie
	}
	# refresh library
	library::routine:reload

	# give access to the buttons
	toolbar::configure .main save -state normal
	toolbar::configure .main routine -state normal
	toolbar::configure .main stage -state normal
	toolbar::configure .main movie -state normal
	toolbar::configure .main property -state normal
	# update state of menu
	.mainmenu.file entryconfigure 4 -state normal
	.mainmenu.file entryconfigure 5 -state normal
	.mainmenu.file entryconfigure 6 -state normal
	.mainmenu.win entryconfigure 1 -state normal
	.mainmenu.win entryconfigure 2 -state normal
	.mainmenu.win entryconfigure 3 -state normal
	.mainmenu.win entryconfigure 5 -state normal
	# remove dialog for waiting
	destroy .wait
}

proc main::new:couple {} {
	open new
	routine::dancer:configure:name .routine 0 Man
	routine::dancer:configure:type .routine 0 Green
	routine::dancer:new .routine 0
	routine::dancer:configure:name .routine 1 Lady
	routine::dancer:configure:sex  .routine 1 woman
	routine::dancer:configure:type .routine 1 Orange
}

#-------------------------------------------------------------------------------------------------

proc main::save {type} {
global status
variable data


	if {$type == "new" || $data(v:filename) == "" || $data(v:firstSave)} {
		set filename [tk_getSaveFile -initialdir "$data(v:currentDir)" \
							-initialfile [file tail $data(v:filename)] \
							-title "Choose name to save the routine"]
		if {$filename == ""} {
			return
		}
		set data(v:currentDir) [file dirname $filename]
		set data(v:filename) $filename
		set data(v:firstSave) 0

		# check if save in the right place
		if {[file dirname $filename] == $status(dir:routine)} {
			tk_messageBox -icon "error" -type ok -default ok -title "Error" \
					-message "You must save the file inside a class !"
			return
		}
		set path [string range $filename 0 [expr [string length $status(dir:routine)]-1]]
#puts "path = $path -- 0 [expr [string length $status(dir:routine)]-1]"
		if {[string compare $path $status(dir:routine)]} {
			tk_messageBox -icon "error" -type ok -message -default ok -title "Error" \
				"You must save the file inside the routine directory.\n($status(dir:routine))"
			set data(v:currentDir) $status(dir:routine)
			return
		}
		# update library
		library::routine:reload
	} else {
		set filename $data(v:filename)
	}

	routine::save .routine $filename
	routine::saved .routine
}

proc main::close {} {
global status
variable data

	# check if current was modified
	if {[winfo exists .routine] && [routine::wasModified .routine]} {
		set dosave [tk_messageBox -icon "question" -type yesno -default yes -title "Error" \
						-message "The routine was modified.\nSave it before ?"]
		if {$dosave == "yes"} {
			save current
		}
	}
	# destroy the windows
	movie::close .movie
	destroy .movie
	stage::close .stage
	destroy .stage
	routine::closeWindow .routine
	destroy .routine
	# update state of toolbar
	toolbar::configure .main save -state disabled
	toolbar::configure .main routine -state disabled
	toolbar::configure .main stage -state disabled
	toolbar::configure .main movie -state disabled
	toolbar::configure .main property -state disabled
	# update state of menu
	.mainmenu.file entryconfigure 4 -state disabled
	.mainmenu.file entryconfigure 5 -state disabled
	.mainmenu.file entryconfigure 6 -state disabled
	.mainmenu.win entryconfigure 1 -state disabled
	.mainmenu.win entryconfigure 2 -state disabled
	.mainmenu.win entryconfigure 3 -state disabled
	.mainmenu.win entryconfigure 5 -state disabled
	# update library
	set data(v:filename) ""
	library::routine:reload
}

#-------------------------------------------------------------------------------------------------

proc main::about {} {
	# create the window
	set top [buildModalDialog "About"]
	# widget
	if {[catch {label $top.img -image imgAbout}]} {
		label $top.img -width 30 -text "\nLogo.gif not found\n" -anchor c \
				-bg lightskyblue -relief groove -bd 2
	}
	label $top.text -text "(c) 1998-2000 Laurent Riesterer\n\nRemarks and bug reports at\nlaurent.riesterer@free.fr"
	pack $top.img $top.text -pady 5
	# wait
	waitforModalDialog
}

proc main::manual {} {
	# create the window
	if {[catch {exec netscape "[file dirname $status(dir:logo.gif)]/manual/index.html" &}]} {
		set top [buildModalDialog "Error"]
		label $top.img -width 30 -text "\nUnable to run Netscape\n" -anchor c \
				-bg lightskyblue -relief groove -bd 2

		label $top.text -text "Open manually the file\n.../gepetto/doc/manual/index.html"
		pack $top.img $top.text -pady 5
	}
	# wait
	waitforModalDialog
}

#-------------------------------------------------------------------------------------------------

proc main::closeWindow {w} {
variable data

	toolbar::check:setOff .main [string range $w 1 end]
}

##################################################################################################
# redefine tkFDialog to remove the "wm transient"

#source ~/tcl/tk8.0/library/tkfbox.tcl

proc tkFDialog {args} {
    global tkPriv
    set w __tk_filedialog
    upvar #0 $w data

    if ![string compare [lindex [info level 0] 0] tk_getOpenFile] {
	set type open
    } else {
	set type save
    }

    tkFDialog_Config $w $type $args

    if {![string compare $data(-parent) .]} {
        set w .$w
    } else {
        set w $data(-parent).$w
    }

    # (re)create the dialog box if necessary
    #
    if {![winfo exists $w]} {
	tkFDialog_Create $w
    } elseif {[string compare [winfo class $w] TkFDialog]} {
	destroy $w
	tkFDialog_Create $w
    }
#    wm transient $w $data(-parent)

    # 5. Initialize the file types menu
    #
    if {$data(-filetypes) != {}} {
	$data(typeMenu) delete 0 end
	foreach type $data(-filetypes) {
	    set title  [lindex $type 0]
	    set filter [lindex $type 1]
	    $data(typeMenu) add command -label $title \
		-command [list tkFDialog_SetFilter $w $type]
	}
	tkFDialog_SetFilter $w [lindex $data(-filetypes) 0]
	$data(typeMenuBtn) config -state normal
	$data(typeMenuLab) config -state normal
    } else {
	set data(filter) "*"
	$data(typeMenuBtn) config -state disabled -takefocus 0
	$data(typeMenuLab) config -state disabled
    }

    tkFDialog_UpdateWhenIdle $w

    # 6. Withdraw the window, then update all the geometry information
    # so we know how big it wants to be, then center the window in the
    # display and de-iconify it.

    wm withdraw $w
    update idletasks
    set x [expr [winfo screenwidth $w]/2 - [winfo reqwidth $w]/2 \
	    - [winfo vrootx [winfo parent $w]]]
    set y [expr [winfo screenheight $w]/2 - [winfo reqheight $w]/2 \
	    - [winfo vrooty [winfo parent $w]]]
    wm geom $w [winfo reqwidth $w]x[winfo reqheight $w]+$x+$y
    wm deiconify $w
    wm title $w $data(-title)

    # 7. Set a grab and claim the focus too.

    set oldFocus [focus]
    set oldGrab [grab current $w]
    if {$oldGrab != ""} {
	set grabStatus [grab status $oldGrab]
    }
    grab $w
    focus $data(ent)
    $data(ent) delete 0 end
    $data(ent) insert 0 $data(selectFile)
    $data(ent) select from 0
    $data(ent) select to   end
    $data(ent) icursor end

    # 8. Wait for the user to respond, then restore the focus and
    # return the index of the selected button.  Restore the focus
    # before deleting the window, since otherwise the window manager
    # may take the focus away so we can't redirect it.  Finally,
    # restore any grab that was in effect.

    tkwait variable tkPriv(selectFilePath)
    catch {focus $oldFocus}
    grab release $w
    wm withdraw $w
    if {$oldGrab != ""} {
	if {$grabStatus == "global"} {
	    grab -global $oldGrab
	} else {
	    grab $oldGrab
	}
    }
    return $tkPriv(selectFilePath)
}

proc tkFDialog_Config {w type argList} {
    upvar #0 $w data

    set data(type) $type

    # 1: the configuration specs
    #
    set specs {
	{-defaultextension "" "" ""}
	{-filetypes "" "" ""}
	{-initialdir "" "" ""}
	{-initialfile "" "" ""}
	{-parent "" "" "."}
	{-title "" "" ""}
    }

    # 2: default values depending on the type of the dialog
    #
    if ![info exists data(selectPath)] {
	# first time the dialog has been popped up
	set data(selectPath) [pwd]
	set data(selectFile) ""
    }

    # 3: parse the arguments
    #
    tclParseConfigSpec $w $specs "" $argList

    if ![string compare $data(-title) ""] {
	if ![string compare $type "open"] {
	    set data(-title) "Open"
	} else {
	    set data(-title) "Save As"
	}
    }

    # 4: set the default directory and selection according to the -initial
    #    settings
    #
    if [string compare $data(-initialdir) ""] {
	if [file isdirectory $data(-initialdir)] {
		#### PATCH to handle space in dir names
	    set data(selectPath) [lindex [glob $data(-initialdir)] 0]
	} else {
	    error "\"$data(-initialdir)\" is not a valid directory"
	}
    }
    set data(selectFile) $data(-initialfile)

    # 5. Parse the -filetypes option
    #
    set data(-filetypes) [tkFDGetFileTypes $data(-filetypes)]

    if ![winfo exists $data(-parent)] {
	error "bad window path name \"$data(-parent)\""
    }
}

