#include "togl.h"

#include "view.h"
#include "body.h"


double SIZE_NECK	= 0.07;
double SIZE_HEAD	= 0.035;
double SIZE_PELVIS	= 0.18;
double SIZE_TORSO	= 0.38;
double SIZE_SHOULDER= 0.19;
double SIZE_ARM		= 0.34;
double SIZE_FOREARM	= 0.30;
double SIZE_HAND	= 0.15;
double SIZE_HIP		= 0.075;
double SIZE_THIGH	= 0.45;
double SIZE_LEG		= 0.40;
double SIZE_FOOT	= 0.045;
double SIZE_TOE		= 0.10;

double HEIGHT_FOOT=0.10061099845916034; 
double HEIGHT_HEAD=0.2078;
double SIZE_HEEL=0.06;


double shift_sh_X=0.0,   shift_sh_Y=0.0, shift_sh_Z=0.0;
double shift_fa_X=0.0,   shift_fa_Y=0.0, shift_fa_Z=0.01;
double shift_wr_X=-0.02, shift_wr_Y=0.0, shift_wr_Z=0.015;

double shift_hp_X=0.0,   shift_hp_Y=0.0, shift_hp_Z=0.0;
double shift_an_X=0.0,   shift_an_Y=0.0, shift_an_Z=0.0;

//double shift_fh_X=0.0,   shift_fh_Y=0.0, shift_fh_Z=0.0;
//double shift_ha_X=0.0,   shift_ha_Y=0.0, shift_ha_Z=0.0;


// Current state during drawing - to prevent unnecessary context changes in OpenGL
int mode;
struct Tcl_Interp *currentInterp;
int selectedPart;
double *selectedPartMatrix;


//================================================================================================

Object::Object()
{
	id = -1;
	red = green = blue = 1.0; 
	mat_specular[0] = mat_specular[1] = mat_specular[2] = 1.0;
	mat_shininess[0] = 30.0;
	selected = FALSE;
	wireframe = FALSE;
}

void Object::Wireframe(struct Tcl_Obj *objv[])
{
	wireframe = TRUE;
}

void Object::Shade(struct Tcl_Obj *objv[])
{
	wireframe = FALSE;
}

void Object::GetShaded(struct Tcl_Obj *objv[])
{
register Tcl_Interp *interp = currentInterp;
extern char *partName[];

	if (!wireframe) {
		Tcl_AppendResult(interp, partName[id-OBJ_PART], " ", NULL);
	}
}

void Object::IsShaded(struct Tcl_Obj *objv[])
{
register Tcl_Interp *interp = currentInterp;
	// to handle multiple answer for limb (e.g. result = 1010110)
	Tcl_AppendResult(interp, wireframe ? "1" : "0", NULL);
}

void Object::Select(struct Tcl_Obj *objv[])
{
	selected = TRUE;
}

void Object::Unselect(struct Tcl_Obj *objv[])
{
	selected = FALSE;
}

void Object::IsSelected(struct Tcl_Obj *objv[])
{
register Tcl_Interp *interp = currentInterp;
	// to handle multiple answer for limb (e.g. result = 1010110)
	Tcl_AppendResult(interp, selected ? "1" : "0", NULL);
}

void Object::GetColor(struct Tcl_Obj *objv[])
{
register Tcl_Interp *interp = currentInterp;
char res[50];

	extern char *partName[];
//	sprintf(res, "%s #%04X%04X%04X ", partName[id-OBJ_PART],
//					(int) (red * 65536), (int) (green * 65536), (int) (blue * 65536));
	sprintf(res, "%s #%02X%02X%02X ", partName[id-OBJ_PART],
					(int) (red * 256), (int) (green * 256), (int) (blue * 256));
	Tcl_AppendResult(interp, res, NULL);
}

void Object::SetColor(struct Tcl_Obj *objv[])
{
register Tcl_Interp *interp = currentInterp;
XColor color;

	if (XParseColor(Tk_Display(Tk_MainWindow(interp)), Tk_Colormap(Tk_MainWindow(interp)),
					  Tcl_GetStringFromObj(objv[0], NULL), &color) == 0) {
		Tcl_SetResult(interp, "conversion from color \"", TCL_STATIC);
		Tcl_AppendResult(interp, Tcl_GetStringFromObj(objv[0], NULL), "\" failed", NULL);
	} else {
		mat_specular[0] = red = (float) color.red / 65536.0;
		mat_specular[1] = green = (float) color.green / 65536.0;
		mat_specular[2] = blue = (float) color.blue / 65536.0;
		mat_shininess[0] = 0.0;
	}
}


void Object::Draw() const
{
#define COLOR 1.0, 1.0, 0.5
static GLfloat sel_specular[] = {COLOR, 0.0};

	if (mode == SELECT) {
		// load the name of the object for picking process
		glLoadName(id);

	} else if (mode == GETPOS) {
		if (id == selectedPart) {
			// get the matrix giving the position of joint
			glGetDoublev(GL_MODELVIEW_MATRIX, selectedPartMatrix);
		}

		// no drawing
		return;

	} else {
		// set up the color/material for drawing
		if (wireframe || (mode == FORCE_WIRE)) {
			if (mode != WIRE && mode != FORCE_WIRE) {
				glDisable(GL_LIGHTING);
				glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
				mode = WIRE;
			}
			if (selected) {
				glColor3f(COLOR);
			} else {
				glColor3f(red, green, blue);
			}
		} else {
			if (mode != SHADED) {
				glEnable(GL_LIGHTING);
				glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
				mode = SHADED;
			}
			if (selected) {
				glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, sel_specular);
			} else {
				glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, mat_specular);
			}
			glMaterialfv(GL_FRONT, GL_SHININESS, mat_shininess);
		}
	}
	glCallList((GLenum) drawId);
#undef COLOR
}

//------------------------------------------------------------------------------------------------
// Drawing of limbs
//------------------------------------------------------------------------------------------------

void Leg::DoOperation(pfnObj op, struct Tcl_Obj *objv[])
{
	thigh.DoOperation(op, objv);
	leg.DoOperation(op, objv);
	foot.DoOperation(op, objv);
	toe.DoOperation(op, objv);
}

void Leg::Draw(const class Body *body) const
{
int side = (leg.GetId()==OBJ_LEFT_LEG ? left : right);

	// Thigh
	glRotated(-body->joints[side+hipX], 1.0, 0.0, 0.0);
	glRotated(body->joints[side+hipZ], 0.0, 0.0, 1.0);
	glRotated(body->joints[side+hipY], 0.0, 1.0, 0.0);
	if (body->ghost != FOOT) thigh.Draw();

	// Leg
	glTranslated(0.0, -SIZE_THIGH, 0.0);
	glRotated(-body->joints[side+kneeX], 1.0, 0.0, 0.0);
	if (body->ghost != FOOT) leg.Draw();

	// Foot
	glTranslated(shift_an_X, shift_an_Y-SIZE_LEG, shift_an_Z);
	glRotated(-body->joints[side+ankleX], 1.0, 0.0, 0.0);
	glRotated(body->joints[side+ankleZ], 0.0, 0.0, 1.0);
	glRotated(body->joints[side+ankleY], 0.0, 1.0, 0.0);
	foot.Draw();
	glTranslated(0.0, -HEIGHT_FOOT, 0.0);

	// End of foot
	glTranslated(0.0, 0.0, SIZE_FOOT);
	glRotated(-body->joints[side+footX], 1.0, 0.0, 0.0);
	toe.Draw();
}


//------------------------------------------------------------------------------------------------

void Arm::DoOperation(pfnObj op, struct Tcl_Obj *objv[])
{
	arm.DoOperation(op, objv);
	forearm.DoOperation(op, objv);
	hand.DoOperation(op, objv);
}

void Arm::Draw(const class Body *body) const
{
int side = (arm.GetId()==OBJ_LEFT_ARM ? left : right);

	// Arm
	glTranslated(0.0, body->joints[side+shoulderTY]/1000.0, 0.0);
	glRotated(body->joints[side+shoulderZ], 0.0, 0.0, 1.0);
	glRotated(-body->joints[side+shoulderX], 1.0, 0.0, 0.0);
	glRotated(body->joints[side+shoulderY], 0.0, 1.0, 0.0);
	if (body->ghost != FOOT) arm.Draw();

	// Forearm
	glTranslated(side==left ? shift_fa_X : -shift_fa_X,
				 shift_fa_Y-SIZE_ARM, shift_fa_Z);
	glRotated(-body->joints[side+elbowX], 1.0, 0.0, 0.0);
	glRotated(-body->joints[side+elbowY], 0.0, 1.0, 0.0);
	if (body->ghost != FOOT) forearm.Draw();

	// Hand
	glTranslated(side==left ? shift_wr_X : -shift_wr_X,
				 shift_wr_Y-SIZE_FOREARM, shift_wr_Z);
	glRotated(body->joints[side+wristZ], 0.0, 0.0, 1.0);
	glRotated(-body->joints[side+wristX], 1.0, 0.0, 0.0);
	glRotated(body->joints[side+wristY], 0.0, 1.0, 0.0);
	if (body->ghost != FOOT) hand.Draw();
}

//------------------------------------------------------------------------------------------------

void Body::Draw(int draw_mode, int jointsOnly, int withRX, int withRY, int withRZ) const
{
	// set up body in the world if needed
	if (!jointsOnly) {
		glTranslated(joints[cogX]/100.0, joints[cogY]/100.0, joints[cogZ]/100.0);
	}

	glTranslated(0.0, Position(pin, 0,0,0, jointsOnly,withRX,withRY,withRZ), 0.0);


	if (withRX) glRotated(joints[cogRX], 1.0, 0.0, 0.0);
	if (withRZ) glRotated(joints[cogRZ], 0.0, 0.0, 1.0);
	if (withRY) glRotated(-joints[cogRY], 0.0, 1.0, 0.0);


	glDisable(GL_LIGHTING);
	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	mode = draw_mode;

	// Pelvis
	glPushMatrix();
	glRotated(joints[pelvisX], 1.0, 0.0, 0.0);
	glRotated(joints[pelvisZ], 0.0, 0.0, 1.0);
	glRotated(joints[pelvisY], 0.0, 1.0, 0.0);
	if (ghost != FOOT) pelvis.Draw();

	glTranslated(0.0, -SIZE_PELVIS, 0.0);

	// Left Leg
	glPushMatrix();
	glTranslated(SIZE_HIP+shift_hp_X, shift_hp_Y, shift_hp_Z);
	glRotated(-joints[pelvisX], 1.0, 0.0, 0.0);
	glRotated(-joints[pelvisZ], 0.0, 0.0, 1.0);
	glRotated(-joints[pelvisY], 0.0, 1.0, 0.0);
	Lleg.Draw(this);
	glPopMatrix();

	// Right Leg
	glPushMatrix();
	glTranslated(-SIZE_HIP-shift_hp_X, shift_hp_Y, shift_hp_Z);
	glRotated(-joints[pelvisX], 1.0, 0.0, 0.0);
	glRotated(-joints[pelvisZ], 0.0, 0.0, 1.0);
	glRotated(-joints[pelvisY], 0.0, 1.0, 0.0);
	Rleg.Draw(this);
	glPopMatrix();

	// Torso
	glPopMatrix();	
	glTranslated(0.0, 0.0, -0.012);
	glRotated(-joints[torsoZ], 0.0, 0.0, 1.0);
	glRotated(joints[torsoX], 1.0, 0.0, 0.0);
	glRotated(joints[torsoY], 0.0, 1.0, 0.0);
	if (ghost != FOOT) torso.Draw();
	glTranslated(0.0, SIZE_TORSO, 0.0);	

	// Left Arm
	glPushMatrix();
	glTranslated(SIZE_SHOULDER+shift_sh_X, shift_sh_Y, shift_sh_Z);
	Larm.Draw(this);
	glPopMatrix();

	// Right Arm
	glPushMatrix();
	glTranslated(-SIZE_SHOULDER+shift_sh_X, shift_sh_Y, shift_sh_Z);
	Rarm.Draw(this);
	glPopMatrix();

	// Head
	glTranslated(0.0, SIZE_NECK, 0.0);
	if (ghost != FOOT) neck.Draw();
	glTranslated(0.0, SIZE_HEAD, 0.0);
	glRotated(joints[headX], 1.0, 0.0, 0.0);
	glRotated(joints[headZ], 0.0, 0.0, 1.0);
	glRotated(joints[headY], 0.0, 1.0, 0.0);

//glTranslated(shift_fh_X, shift_fh_Y, shift_fh_Z);
	glTranslated(0.0, -0.02, 0.032);
	if (ghost != FOOT) headFace.Draw();
//glTranslated(shift_ha_X, shift_ha_Y, shift_ha_Z);
	glTranslated(0.0, 0.05, -0.06);
	if (ghost != FOOT) headHair.Draw();
}


double Body::Position(int what, double *x, double *y, double *z, 
					  int jointsOnly, int withRX, int withRY, int withRZ) const
{
double m[4][4], ly = 0.0;

#define getY	glGetDoublev(GL_MODELVIEW_MATRIX,&m[0][0]); if (m[3][1]<ly) ly=m[3][1]
#define getAll	glGetDoublev(GL_MODELVIEW_MATRIX,&m[0][0]); *x=m[3][0]; *y=m[3][1]; *z=m[3][2]

//#define getY	glGetDoublev(GL_MODELVIEW_MATRIX,&m[0][0]); fprintf(stderr, "%f vs %f\n", m[3][1], ly); if (m[3][1]<ly) ly=m[3][1]
//fprintf(stderr, "---------------\n");
//fprintf(stderr, "Position  %d %d %d %d  --  ", jointsOnly,withRX,withRY,withRZ);
//extern char *jointName[];
//fprintf(stderr, "pin=%s  --  what=%s\n", what==pin?"true":"false", jointName[what]);

	// save state
	glPushMatrix();

	glLoadIdentity();
	// position of body
	if (what != pin) glTranslated(0.0,Position(pin, 0,0,0, jointsOnly,withRX,withRY,withRZ),0.0);
	if (what != pin && !jointsOnly) {
		glTranslated(joints[cogX]/100.0, joints[cogY]/100.0, joints[cogZ]/100.0);
	}
	if (withRX || what==pin) glRotated(joints[cogRX], 1.0, 0.0, 0.0);
	if (withRZ || what==pin) glRotated(joints[cogRZ], 0.0, 0.0, 1.0);
	if (withRY || what==pin) glRotated(-joints[cogRY], 0.0, 1.0, 0.0);

	// draw pelvis
	glPushMatrix();
	glRotated(joints[pelvisX], 1.0, 0.0, 0.0);
	glRotated(joints[pelvisZ], 0.0, 0.0, 1.0);
	glRotated(joints[pelvisY], 0.0, 1.0, 0.0);
	glTranslated(0.0, -SIZE_PELVIS, 0.0);
	if (what == bPelvis) {
		getAll;
		goto done;
	}

	// Left Leg
//fprintf(stderr, "lleg:");
	glPushMatrix();
		glTranslated(SIZE_HIP+shift_hp_X, shift_hp_Y, shift_hp_Z);
		if (what == lHip) {
			getAll;
			goto done;
		}
		getY;
		glRotated(-joints[pelvisX], 1.0, 0.0, 0.0);
		glRotated(-joints[pelvisZ], 0.0, 0.0, 1.0);
		glRotated(-joints[pelvisY], 0.0, 1.0, 0.0);
		glRotated(-joints[lHipX], 1.0, 0.0, 0.0);
		glRotated(joints[lHipZ], 0.0, 0.0, 1.0);
		glRotated(joints[lHipY], 0.0, 1.0, 0.0);
		// Leg
		glTranslated(0.0, -SIZE_THIGH, 0.0);
		if (what == lKnee) {
			getAll;
			goto done;
		}
		getY;
		glRotated(-joints[lKneeX], 1.0, 0.0, 0.0);
		// Foot
		glTranslated(shift_an_X, shift_an_Y-SIZE_LEG, shift_an_Z);
		if (what == lAnkle) {
			getAll;
			goto done;
		}
		getY;
		glRotated(-joints[lAnkleX], 1.0, 0.0, 0.0);
		glRotated(joints[lAnkleZ], 0.0, 0.0, 1.0);
		glRotated(joints[lAnkleY], 0.0, 1.0, 0.0);
		glTranslated(0.0, -HEIGHT_FOOT, 0.0);
		if (what == pin) {
			glTranslated(0.0, 0.0, -SIZE_HEEL);
			getY;
			glTranslated(0.0, 0.0, +SIZE_HEEL);
		}
		// End of foot
		glTranslated(0.0, 0.0, SIZE_FOOT);
		if (what == lFoot) {
			getAll;
			goto done;
		}
		getY;
		glRotated(-joints[lFootX], 1.0, 0.0, 0.0);
		glTranslated(0.0, 0.0, SIZE_TOE);
        getY;
	glPopMatrix();

	// Right Leg
//fprintf(stderr, "rleg:");
	glPushMatrix();
		glTranslated(-SIZE_HIP-shift_hp_X, shift_hp_Y, shift_hp_Z);
		if (what == rHip) {
			getAll;
			goto done;
		}
		getY;
		glRotated(-joints[pelvisX], 1.0, 0.0, 0.0);
		glRotated(-joints[pelvisZ], 0.0, 0.0, 1.0);
		glRotated(-joints[pelvisY], 0.0, 1.0, 0.0);
		glRotated(-joints[rHipX], 1.0, 0.0, 0.0);
		glRotated(joints[rHipZ], 0.0, 0.0, 1.0);
		glRotated(joints[rHipY], 0.0, 1.0, 0.0);
		// Leg
		glTranslated(0.0, -SIZE_THIGH, 0.0);
		if (what == rKnee) {
			getAll;
			goto done;
		}
		getY;
		glRotated(-joints[rKneeX], 1.0, 0.0, 0.0);
		// Foot
		glTranslated(shift_an_X, shift_an_Y-SIZE_LEG, shift_an_Z);
		if (what == rAnkle) {
			getAll;
			goto done;
		}
		getY;
		glRotated(-joints[rAnkleX], 1.0, 0.0, 0.0);
		glRotated(joints[rAnkleZ], 0.0, 0.0, 1.0);
		glRotated(joints[rAnkleY], 0.0, 1.0, 0.0);
		glTranslated(0.0, -HEIGHT_FOOT, 0.0);
		if (what == pin) {
			glTranslated(0.0, 0.0, -SIZE_HEEL);
			getY;
			glTranslated(0.0, 0.0, +SIZE_HEEL);
		}
		// End of foot
		glTranslated(0.0, 0.0, SIZE_FOOT);
		if (what == rFoot) {
			getAll;
			goto done;
		}
		getY;
		glRotated(-joints[rFootX], 1.0, 0.0, 0.0);
		glTranslated(0.0, 0.0, SIZE_TOE);
        getY;
	glPopMatrix();

	// Torso
//fprintf(stderr, "torso:");
	glPopMatrix();
	glTranslated(0.0, 0.0, -0.012);
	if (what == bTorso) {
		getAll;
		goto done;
	}
	getY;
	glRotated(-joints[torsoZ], 0.0, 0.0, 1.0);
	glRotated(joints[torsoX], 1.0, 0.0, 0.0);
	glRotated(joints[torsoY], 0.0, 1.0, 0.0);
	glTranslated(0.0, SIZE_TORSO, 0.0);	

	// Left Arm
//fprintf(stderr, "larm:");
	glPushMatrix();
		glTranslated(SIZE_SHOULDER+shift_sh_X, shift_sh_Y, shift_sh_Z);
		glTranslated(0.0, joints[lShoulderTY]/1000.0, 0.0);
		if (what == lShoulder) {
			getAll;
			goto done;
		}
		getY;
		glRotated(joints[lShoulderZ], 0.0, 0.0, 1.0);
		glRotated(-joints[lShoulderX], 1.0, 0.0, 0.0);
		glRotated(joints[lShoulderY], 0.0, 1.0, 0.0);
		// Forearm
		glTranslated(shift_fa_X, shift_fa_Y-SIZE_ARM, shift_fa_Z);
		if (what == lElbow) {
			getAll;
			goto done;
		}
		getY;
		glRotated(-joints[lElbowX], 1.0, 0.0, 0.0);
		glRotated(-joints[lElbowY], 0.0, 1.0, 0.0);
		// Hand
		glTranslated(shift_wr_X, shift_wr_Y-SIZE_FOREARM, shift_wr_Z);
		if (what == lWrist) {
			getAll;
			goto done;
		}
		getY;
		glRotated(joints[lWristZ], 0.0, 0.0, 1.0);
		glRotated(-joints[lWristX], 1.0, 0.0, 0.0);
		glRotated(joints[lWristY], 0.0, 1.0, 0.0);
		glTranslated(0.0, -SIZE_HAND, 0.0);
		getY;
	glPopMatrix();

	// Right Arm
//fprintf(stderr, "rarm:");
	glPushMatrix();
		glTranslated(-SIZE_SHOULDER+shift_sh_X, shift_sh_Y, shift_sh_Z);
		glTranslated(0.0, joints[rShoulderTY]/1000.0, 0.0);
		if (what == rShoulder) {
			getAll;
			goto done;
		}
		getY;
		glRotated(joints[rShoulderZ], 0.0, 0.0, 1.0);
		glRotated(-joints[rShoulderX], 1.0, 0.0, 0.0);
		glRotated(joints[rShoulderY], 0.0, 1.0, 0.0);
		// Forearm
		glTranslated(shift_fa_X, shift_fa_Y-SIZE_ARM, shift_fa_Z);
		if (what == rElbow) {
			getAll;
			goto done;
		}
		getY;
		glRotated(-joints[rElbowX], 1.0, 0.0, 0.0);
		glRotated(-joints[rElbowY], 0.0, 1.0, 0.0);
		// Hand
		glTranslated(shift_wr_X, shift_wr_Y-SIZE_FOREARM, shift_wr_Z);
		if (what == rWrist) {
			getAll;
			goto done;
		}
		getY;
		glRotated(joints[rWristZ], 0.0, 0.0, 1.0);
		glRotated(-joints[rWristX], 1.0, 0.0, 0.0);
		glRotated(joints[rWristY], 0.0, 1.0, 0.0);
		glTranslated(0.0, -SIZE_HAND, 0.0);
		getY;
	glPopMatrix();

	// Head
//fprintf(stderr, "head:");
	glTranslated(0.0, SIZE_NECK, 0.0);
	if (what == bHead || what == bNeck) {
		getAll;
		goto done;
	}
	glRotated(joints[headX], 1.0, 0.0, 0.0);
	glRotated(joints[headZ], 0.0, 0.0, 1.0);
	glRotated(joints[headY], 0.0, 1.0, 0.0);
	glTranslated(0.0, SIZE_HEAD+HEIGHT_HEAD, 0.0);
	getY;

	// restore state
done:
	glPopMatrix();

//fprintf(stderr, "result = %f\n", ly);
	return -ly;
}


//================================================================================================


void Floor::Draw() const
{
	glBegin(GL_LINES);
	glColor3dv(color);
	for(int x=-nbX; x<=nbX; x++) {
		if (x != 0) {
			glVertex3d(x*padX, 0.0, -nbZ*padZ);
			glVertex3d(x*padX, 0.0, +nbZ*padZ);
		}
	}
	for(int z=-nbZ; z<=nbZ; z++) {
		if (z != 0) {
			glVertex3d(-nbX*padX, 0.0, z*padZ);
			glVertex3d(+nbX*padX, 0.0, z*padZ);
		}
	}
	glEnd();
	glBegin(GL_LINES);
	glColor3dv(colorAxe);
	  glVertex3d(0.0, 0.0, -nbZ*padZ);
	  glVertex3d(0.0, 0.0, +nbZ*padZ);

	  glVertex3d(-nbX*padX, 0.0, 0.0);
	  glVertex3d(+nbX*padX, 0.0, 0.0);
	glEnd();
}
