#include "togl.h"
#include "view.h"

/*
 * The following variable is a special hack that is needed in order for
 * Sun shared libraries to be used for Tcl.
 */

#ifdef TK_TEST
extern int		Tktest_Init _ANSI_ARGS_((Tcl_Interp *interp));
#endif /* TK_TEST */

extern "C" int Tix_Init(Tcl_Interp *interp);
extern void Togl_Setup(Tcl_Interp *interp);
extern void Gepetto_Setup(Tcl_Interp *interp);

extern "C" { int uncompress (unsigned char *dest, unsigned int *destLen,
							 unsigned char *source, unsigned int sourceLen); }

/*
 *----------------------------------------------------------------------
 *
 * main --
 *
 *	This is the main program for the application.
 *
 * Results:
 *	None: Tk_Main never returns here, so this procedure never
 *	returns either.
 *
 * Side effects:
 *	Whatever the application does.
 *
 *----------------------------------------------------------------------
 */

int main(int argc, char **argv)
{
    Tk_Main(argc, argv, Tcl_AppInit);
    return 0;			/* Needed only to prevent compiler warning. */
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_AppInit --
 *
 *	This procedure performs application-specific initialization.
 *	Most applications, especially those that incorporate additional
 *	packages, will have their own version of this procedure.
 *
 * Results:
 *	Returns a standard Tcl completion code, and leaves an error
 *	message in interp->result if an error occurs.
 *
 * Side effects:
 *	Depends on the startup script.
 *
 *----------------------------------------------------------------------
 */

int Tcl_AppInit(Tcl_Interp *interp)
{
    if (Tcl_Init(interp) == TCL_ERROR) {
		return TCL_ERROR;
    }
    if (Tk_Init(interp) == TCL_ERROR) {
		return TCL_ERROR;
    }
    Tcl_StaticPackage(interp, "Tk", Tk_Init, Tk_SafeInit);

	if (Tix_Init(interp) == TCL_ERROR) {
		return TCL_ERROR;
	}
	if (Togl_Init(interp) == TCL_ERROR) {
		return TCL_ERROR;
	}

	Togl_Setup(interp);
	Gepetto_Setup(interp);


    Tcl_CreateCommand(interp, "body", BodyCreate_Cmd,
                      (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "light", LightCreate_Cmd,
                      (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "routine", RoutineCreate_Cmd,
                      (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

    Tcl_CreateObjCommand(interp, "check", Check_ObjCmd,
                      (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);


	// include the script (source of tcl + pixmap = self-contained)
	extern unsigned int scriptSize;
	extern unsigned int scriptCompressedSize;
	extern unsigned char script[];
	unsigned char *tmp = new unsigned char[scriptSize+10];
	unsigned int uncompressed = scriptSize+5;
	if (uncompress(tmp, &uncompressed, script, scriptCompressedSize) != 0) {
		fprintf(stderr, "ERROR in initialisation\n");
		return TCL_ERROR;
	}
	if (uncompressed != scriptSize) {
		fprintf(stderr, "ERROR in initialisation\n");
		return TCL_ERROR;
	}
	// eval the library of scripts
	int result = Tcl_Eval(interp, (char *) tmp);
	if (result != TCL_OK) {
		fprintf(stderr, "ERROR in initialisation :\n%s\n", Tcl_GetStringResult(interp));
		return TCL_ERROR;
	}
	// free tmp
	delete[] tmp;

	return TCL_OK;
}
