#include "togl.h"
#include <math.h>
#include <GL/glu.h>

#include "view.h"


// The following variable is a special hack that is needed in order for
// Sun shared libraries to be used for Tcl.
#ifndef WIN32
extern "C" int matherr();
int *tclDummyMathPtr = (int *) matherr;
#endif


//================================================================================================

void OpenGL_Setup()
{
	// colored axis
	glNewList(OBJ_AXIS, GL_COMPILE);
	  glBegin(GL_LINES);
      glColor3f(1.0, 0.0, 0.0);			// X axis -- red
      glVertex3f(0.0, 0.0, 0.0);
      glVertex3f(1.0, 0.0, 0.0);
      glColor3f(0.0, 1.0, 0.0);			// Y axis -- green
      glVertex3f(0.0, 0.0, 0.0);
      glVertex3f(0.0, 1.0, 0.0);
      glColor3f(0.0, 0.0, 1.0);			// Z axis -- blue
      glVertex3f(0.0, 0.0, 0.0);
	  glVertex3f(0.0, 0.0, 1.0);
	  glEnd();
	glEndList();

	// spheres
	glNewList(OBJ_SPHERE, GL_COMPILE);
	  GLUquadricObj *s = gluNewQuadric();
	  gluQuadricDrawStyle(s, (GLenum) GLU_LINE);
	  glRotatef(90.0, 1.0, 0.0, 0.0);
	  gluSphere(s, 1, 9, 9);
	glEndList();

	glNewList(OBJ_SPHERE_XZ, GL_COMPILE);
	  glBegin(GL_LINES);
		glVertex3f(-5.0, 0.0, 0.0);
		glVertex3f(+5.0, 0.0, 0.0);
		glVertex3f(0.0, -5.0, 0.0);
		glVertex3f(0.0, +5.0, 0.0);
		glVertex3f(0.0, 0.0, -5.0);
		glVertex3f(0.0, 0.0, +5.0);
	  glEnd();
	  glCallList(OBJ_SPHERE);
	glEndList();

	glNewList(OBJ_SPHERE_X_ONLY, GL_COMPILE);
	  glBegin(GL_LINES);
		glVertex3f(-5.0, 0.0, 0.0);
		glVertex3f(+5.0, 0.0, 0.0);
		glVertex3f(0.0, -5.0, 0.0);
		glVertex3f(0.0, +5.0, 0.0);
	  glEnd();
	  glCallList(OBJ_SPHERE);
	glEndList();

	glNewList(OBJ_SPHERE_Z_ONLY, GL_COMPILE);
	  glBegin(GL_LINES);
		glVertex3f(0.0, -5.0, 0.0);
		glVertex3f(0.0, +5.0, 0.0);
		glVertex3f(0.0, 0.0, -5.0);
		glVertex3f(0.0, 0.0, +5.0);
	  glEnd();
	  glCallList(OBJ_SPHERE);
	glEndList();

	// disk
	glNewList(OBJ_DISK, GL_COMPILE);
	  glBegin(GL_LINES);
		glVertex3f(-5.0, 0.0, 0.0);
		glVertex3f(+5.0, 0.0, 0.0);
		glVertex3f(0.0, -5.0, 0.0);
		glVertex3f(0.0, +5.0, 0.0);
		glVertex3f(0.0, 0.0, -5.0);
		glVertex3f(0.0, 0.0, +5.0);
	  glEnd();
	  GLUquadricObj *d = gluNewQuadric();
	  gluQuadricDrawStyle(d, (GLenum) GLU_LINE);
	  glRotatef(90.0, 1.0, 0.0, 0.0);
	  gluDisk(s, 1, 1.25, 36, 1);
	glEndList();

	ObjectsInit();
}

//------------------------------------------------------------------------------------------------
// Callback used by the OpenGL windows

void OpenGL_Create(struct Togl *togl)
{
static int setup = FALSE;

    glDepthFunc(GL_LESS);

#ifndef WIN32
	if (!setup) {
#endif
		OpenGL_Setup();
#ifndef WIN32
		setup = TRUE;
	}
#endif

	// create the appropriate structure
	if (strcmp(togl->Class, "View3D")==0) {
		togl->Client_Data = (ClientData) new View3D(togl);
	} else if (strcmp(togl->Class, "ViewMovie")==0) {
		togl->Client_Data = (ClientData) new ViewMovie(togl);
	} else if (strcmp(togl->Class, "ViewTrack")==0) {
		togl->Client_Data = (ClientData) new ViewTrack(togl);
	} else {
		togl->Client_Data = (ClientData) new View3D(togl);
	}

	// force redisplay
	Togl_PostRedisplay(togl);
}

void OpenGL_Destroy(struct Togl *togl)
{
class AbstractView *view = (class AbstractView *) Togl_GetClientData(togl);

	if (view != NULL) {
		delete view;
	}
}

void OpenGL_Display(struct Togl *togl)
{
class AbstractView *view = (class AbstractView *) Togl_GetClientData(togl);

	if (view != NULL) {
		view->Redisplay();
	}
	Togl_SwapBuffers(togl);
}

//================================================================================================

void Togl_Setup(Tcl_Interp *interp)
{
	// Register class of widget
	Togl_RegisterClass("View3D", View3D_ObjCmd);
	Togl_RegisterClass("ViewMovie", ViewMovie_ObjCmd);
	Togl_RegisterClass("ViewTrack", ViewTrack_ObjCmd);

	// Specify the C callback functions for widget creation, display, and destruction.
	Togl_DisplayFunc(OpenGL_Display);
	Togl_CreateFunc(OpenGL_Create);
	Togl_DestroyFunc(OpenGL_Destroy);

	// Link variables
#define vard(nameTcl, nameC) {char tmp[100]; strcpy(tmp, nameTcl); \
							 Tcl_LinkVar(interp, tmp, (char *) &(nameC), TCL_LINK_DOUBLE);}

/*extern double SIZE_NECK;
extern double SIZE_HEAD;
extern double SIZE_PELVIS;
extern double SIZE_TORSO;
extern double SIZE_SHOULDER;
extern double SIZE_ARM;
extern double SIZE_FOREARM;
extern double SIZE_HAND;
extern double SIZE_HIP;
extern double SIZE_THIGH;
extern double SIZE_LEG;
extern double SIZE_FOOT;
extern double SIZE_TOE;

	vard("s(neck)", SIZE_NECK)
	vard("s(head)", SIZE_HEAD)
	vard("s(pelvis)", SIZE_PELVIS)
	vard("s(torso)", SIZE_TORSO)
	vard("s(shoulder)", SIZE_SHOULDER)
	vard("s(arm)", SIZE_ARM)
	vard("s(forearm)", SIZE_FOREARM)
	vard("s(hand)", SIZE_HAND)
	vard("s(hip)", SIZE_HIP)
	vard("s(thigh)", SIZE_THIGH)
	vard("s(leg)", SIZE_LEG)
	vard("s(foot)", SIZE_FOOT)
	vard("s(toe)", SIZE_TOE)*/
/*extern double totoX, totoY, totoZ;
	vard("s(totoX)", totoX)
	vard("s(totoY)", totoY)
	vard("s(totoZ)", totoZ)*/

/*extern double shift_an_X, shift_an_Y, shift_an_Z;
extern double shift_fa_X, shift_fa_Y, shift_fa_Z;
extern double shift_wr_X, shift_wr_Y, shift_wr_Z;
extern double shift_sh_X, shift_sh_Y, shift_sh_Z;
extern double shift_hp_X, shift_hp_Y, shift_hp_Z;
	vard("anX", shift_an_X);
	vard("anY", shift_an_Y);
	vard("anZ", shift_an_Z);
	vard("faX", shift_fa_X);
	vard("faY", shift_fa_Y);
	vard("faZ", shift_fa_Z);
	vard("wrX", shift_wr_X);
	vard("wrY", shift_wr_Y);
	vard("wrZ", shift_wr_Z);
	vard("hpX", shift_hp_X);
	vard("hpY", shift_hp_Y);
	vard("hpZ", shift_hp_Z);
	vard("shX", shift_sh_X);
	vard("shY", shift_sh_Y);
	vard("shZ", shift_sh_Z);*/

/*extern double shift_ha_X, shift_ha_Y, shift_ha_Z, shift_fh_X, shift_fh_Y, shift_fh_Z;
	vard("fhX", shift_fh_X);
	vard("fhY", shift_fh_Y);
	vard("fhZ", shift_fh_Z);
	vard("haX", shift_ha_X);
	vard("haY", shift_ha_Y);
	vard("haZ", shift_ha_Z);*/
}

//------------------------------------------------------------------------------------------------
// Register the dashed & dotted line used by the routine editor

void Gepetto_Setup(Tcl_Interp *interp)
{
static char dash_bits[] = {1,1,1,1,0,0,0,0,1,1,1,1,0,0,0,0};
static char dot_bits[]  = {1,1,0,0,1,1,0,0,1,1,0,0,1,1,0,0};


	Tk_DefineBitmap(interp, Tk_GetUid("linedash"), dash_bits, 1, 16);
	Tk_DefineBitmap(interp, Tk_GetUid("linedot"), dot_bits, 1, 16);
}
