#include <stdio.h>
#include <stdlib.h>
#include <string.h>


FILE *in, *out;


#define MAX					200
#define MAX_ELEMENTS		20
#define MAX_NESTED_LISTS		5


char *class, *codeSet = "/* nothing */", *codeGet = "/* nothing */";
char *extension = NULL;
char *name[MAX];
char *var[MAX];
char *codeVarSet[MAX];
char *userSet[MAX];
char *userGet[MAX];
char *listVar[MAX];
int  list[MAX][MAX_ELEMENTS];
enum {INT=0, DOUBLE, BOOL, STRING, COLOR, USER, LIST} type[MAX];


char *funcSet[] = {"Tcl_GetIntFromObj", "Tcl_GetDoubleFromObj",
				   "Tcl_GetBooleanFromObj", NULL, NULL};
char *funcGet[] = {"Tcl_NewIntObj", "Tcl_NewDoubleObj",
				   "Tcl_NewBooleanObj", "Tcl_NewStringObj", NULL};


//================================================================================================

#define get_code	strcpy(code, tmp+n); strcpy(tmp, code); \
					while (tmp[strlen(tmp)-2] == '\\') { \
						code[strlen(code)-2] = '\0'; \
						fgets(tmp, 200, in); \
						strcat(code,"\n"); \
						strcat(code,tmp); \
					}

#define find(msg)	for(j=0; j<i; j++) { \
						if (strcmp(name[j],p1)==0) \
							break; \
					} \
					if (j==i) { \
						fprintf(stderr, "[%s] " msg " bad variable '%s'\n", argv[1], p1); \
						exit(1); \
					}



//================================================================================================
//
//	...::Set()
//
//================================================================================================


void OutputSet(int idx, int level, char *obj)
{
int i;
char tmp[30], *ptr, *pos, save;

	if (level > MAX_NESTED_LISTS) {
		fprintf(stderr, "Too many nested list\n");
		exit(1);
	}

	if (type[idx]==USER) {
		// User defined
		fprintf(out,"%s\n", userSet[idx]);

	} else if (type[idx] == LIST) {
		// List of variables
 		fprintf(out,"Tcl_ListObjGetElements(interp, %s, &__nb, &__list[%d]);\n", obj, level);
		if (level==0) {
			fprintf(out,"if (__nb<%d) {"
					"    Tcl_WrongNumArgs(interp, 1, &objv[2], \"{%s}\");"
					"    return TCL_ERROR;"
					"}", list[idx][0], listVar[idx]);
		} else {
			fprintf(out,"if (__nb<%d) {"
					"    __tmp = Tcl_NewStringObj(\"%s\",-1);"
					"    Tcl_WrongNumArgs(interp, 1, &__tmp, \"{%s}\");"
					"    return TCL_ERROR;"
					"}", list[idx][0], name[idx], listVar[idx]);
		}
		for(i=1; i<=list[idx][0]; i++) {
			sprintf(tmp, "__list[%d][%d]", level, i-1);
			OutputSet(list[idx][i], level+1, tmp); 
		}
	} else {
		// Basic type
		if (type[idx] == STRING) {
			// String
			fprintf(out,"if (%s != NULL) delete %s;", var[idx], var[idx]);
			fprintf(out,"__str = Tcl_GetStringFromObj(%s, NULL);"
						"%s = new char[strlen(__str)];"
						"strcpy(%s, __str);", obj, var[idx], var[idx]);
		} else if (type[idx] == COLOR) {
			// Color
			fprintf(out,"if (XParseColor(Tk_Display(Tk_MainWindow(interp)),"
						"				Tk_Colormap(Tk_MainWindow(interp)),"
						"				Tcl_GetStringFromObj(%s, NULL), &__color) == 0) {"
						"	Tcl_SetResult(interp, \"conversion from color \\\"\", TCL_STATIC);"
						"	Tcl_AppendResult(interp, Tcl_GetStringFromObj(objv[3], NULL),"
						"			\"\\\" can't be done\", NULL);"
						"} else {"
//"fprintf(stderr, \"(%%04X,%%04X,%%04X)\n\", _color.red,_color.green,_color.blue);"
						"	%s[0] = (double) __color.red / 65536.0;"
						"	%s[1] = (double) __color.green / 65536.0;"
						"	%s[2] = (double) __color.blue / 65536.0;"
						"	%s[3] = 0.0;"
						"}", obj, var[idx], var[idx], var[idx], var[idx]);
		} else {
			// Int, double, boolean
			fprintf(out,"%s(NULL, %s, &(%s));", funcSet[type[idx]], obj,
					var[idx]);
		}

		// Special user code for given variable
		pos = codeVarSet[idx];
		while ((ptr = index(pos, '%')) != NULL) {
			save = ptr[0]; ptr[0] = '\0'; fprintf(out,pos); ptr[0] = save;
			if (ptr[1] == '!') {
				// %! == index of var
				fprintf(out,"%d",idx);
				pos = ptr+2;
			} else if (ptr[1] == '$') {
				// %$ == var name
				fprintf(out,"%s",var[idx]);
				pos = ptr+2;
			} else {
				fprintf(out,"%%");
				pos = ptr+1;
			}
		}
		fprintf(out, pos);
	}
}

//================================================================================================
//
//	...::Get()
//
//================================================================================================

void OutputGet(int idx, int level, char *proto, char *epilog)
{
int i;
char tmp[200];

	if (type[idx]==USER) {
		// User defined
		fprintf(out,"%s", userGet[idx]);
	} else if (type[idx] == LIST) {
		// List of variables
		fprintf(out,"__list[%d] = Tcl_NewListObj(0, NULL);", level);
		for(i=1; i<=list[idx][0]; i++) {
			sprintf(tmp, "Tcl_ListObjAppendElement(interp,__list[%d],", level);
			OutputGet(list[idx][i], level+1, tmp, ")");
		}
		if (level == 0) {
			fprintf(out,"__res = __list[0];");
		} else {
			fprintf(out,"Tcl_ListObjAppendElement(interp,__list[%d],__list[%d]);",
					level-1, level);
		}
	} else {
		// Basic types
		if (type[idx] == COLOR) {
			// Color
			fprintf(out,"sprintf(__string, \"#%%04X%%04X%%04X\","
						"			(int) (%s[0] * 65535.0), (int) (%s[1] * 65535.0),"
						"			(int) (%s[1] * 65535.0));",
						var[idx], var[idx], var[idx]);
			fprintf(out,"%sTcl_NewStringObj(__string, -1)%s;", proto, epilog);
		} else {
			// Int, double, boolean, string
			fprintf(out,"%s%s(%s%s)%s;", proto, funcGet[type[idx]], var[idx], 
					type[idx]==STRING ? ",-1" : "", epilog);
		}
	}
}


//================================================================================================
//
//		M A I N    E N T R Y
//
//================================================================================================

int main(int argc, char **argv)
{
char tmp[10000], code[10000], *ptr;
char p1[200],p2[200],p3[200];
int i,j,k,n;


	if (argc<2) {
		fprintf(stderr, "Usage: %s file\n", argv[0]);
		exit(1);
	}


	if ((in = fopen(argv[1], "r")) == NULL) {
		fprintf(stderr, "Cannot open input\n");
		exit(1);
	}

	sprintf(tmp, "%scc", argv[1]);
	if ((out = fopen(tmp, "w")) == NULL) {
		fprintf(stderr, "Cannot open output\n");
		exit(1);
	}


	i = 0;
	while (fgets(tmp, 200, in)) {
		if (*tmp == '\n' || *tmp == '#')
			continue;


		// class name
		if (sscanf(tmp, "CLASS %s", p1)==1) {
			class = strdup(p1);
			continue;
		}

		// extension of class name
		if (sscanf(tmp, "EXTENDS %s", p1)==1) {
			extension = strdup(p1);
			continue;
		}

		// code set
		if (sscanf(tmp, "GLOBAL %n%s", &n, p1)==1) {
			get_code;
			fprintf(out, code);
			continue;
		}

		// code set
		if (sscanf(tmp, "CODE_SET %n%s", &n, p1)==1) {
			get_code;
			codeSet = strdup(code);
			continue;
		}
		// code get
		if (sscanf(tmp, "CODE_GET %n%s", &n, p1)==1) {
			get_code;
			codeGet = strdup(code);
			continue;
		}

		// special code
		if (sscanf(tmp, "USER_SET %s%n", p1, &n)==1) {
			find("USER_SET")
			get_code;
			userSet[j] = strdup(code);
			continue;
		}
		if (sscanf(tmp, "USER_GET %s%n", p1, &n)==1) {
			find("USER_GET")
			get_code;
			userGet[j] = strdup(code);
			continue;
		}

		// list of vars
		if (sscanf(tmp, "LIST %s%n", p1, &n)==1) {
			name[i] = strdup(p1);
			type[i] = LIST;
			k = 1;

			get_code;
			ptr = code;

			listVar[i] = (char *) malloc(5000);
			listVar[i][0] = '\0';

			while(sscanf(ptr, "%s%n", p1, &n)==1) {
				find("LIST")
				if (type[j] != LIST) {
/*					// remove listName: in var of type listName:var
					ptrL = name[i];		// name of list obj as prefix
					ptrN = name[j];		// name of the var
					if (index(ptrN, ':')) {
						while (*ptrN == *ptrL) {
							ptrN++;
							ptrL++;
						}
						if (*ptrN == ':') ptrN++;
					}
					strcat(listVar[i], ptrN);*/
					strcat(listVar[i], name[j]);
				} else {
					strcat(listVar[i], "{");
					strcat(listVar[i], listVar[j]);
					strcat(listVar[i], "}");
				}
				strcat(listVar[i], " ");
				list[i][k++] = j;
				ptr += n;
			}
			listVar[i][strlen(listVar[i])-1] = '\0'; // remove last space
			list[i][0] = k-1;
			i++;
			continue;
		}

		// var definition
		if (sscanf(tmp, "%s%s%s%n", p1,p2,p3,&n) != 3) {
			fprintf(stderr, "[%s] bad format '%s'\n", argv[1], tmp);
			exit(1);
		}

		name[i] = strdup(p1);
		var[i] = strdup(p3);
		if (strcmp(p2, "int")==0) {
			type[i] = INT;
		} else if (strcmp(p2, "double")==0) {
			type[i] = DOUBLE;
		} else if (strcmp(p2, "bool")==0) {
			type[i] = BOOL;
		} else if (strcmp(p2, "string")==0) {
			type[i] = STRING;
		} else if (strcmp(p2, "color")==0) {
			type[i] = COLOR;
		} else if (strcmp(p2, "user")==0) {
			type[i] = USER;
			userSet[i] = userGet[i] = "";
		}
		get_code;
		codeVarSet[i] = code[0]=='\n'?"":strdup(code);
		i++;
	}

	// common part
	fprintf(out, "static char *setgetPtr[] = {\n");
	for(j=0; j<i; j++) {
		fprintf(out, "\"%s\"%s", name[j], j<i-1?",":", (char *) NULL ");
	}
	fprintf(out, "};\n\n");


	// generate Set
	fprintf(out,"int %s::Set(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) {\n"
				"int __i, __nb;"
				"char *__str;"
				"Tcl_Obj **__list[%d], *__tmp;"
				"XColor __color;", class, MAX_NESTED_LISTS);

	if (extension) {
		fprintf(out,"    if (%s::Set(interp, objc, objv) == TCL_OK) return TCL_OK;", extension);
		fprintf(out,"    Tcl_ResetResult(interp);");
	}

	fprintf(out,"	 if (objc < 4) {"
                "		 Tcl_WrongNumArgs(interp, 2, objv, \"variable value\");"
				"        return TCL_ERROR;"
				"    }"
				"    if (Tcl_GetIndexFromObj(interp, objv[2],"
				"							 setgetPtr, \"variable\", 0, &__i) != TCL_OK) {"
				"    	return TCL_ERROR;"
				"    }"
				"    switch (__i) {\n");

	for(j=0; j<i; j++) {
		fprintf(out,"case %2d: ", j);
		OutputSet(j, 0, "objv[3]");
		fprintf(out,"break;");
	}
	fprintf(out,"}\n%s\n", codeSet);
	fprintf(out,"return TCL_OK;\n}\n\n");


	// generate Get
	fprintf(out,"int %s::Get(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) const {\n"
				"int __i;"
				"char __string[100];"
				"Tcl_Obj *__res = NULL, *__list[%d];", class, MAX_NESTED_LISTS);

	if (extension) {
		fprintf(out,"    if (%s::Get(interp, objc, objv) == TCL_OK) return TCL_OK;", extension);
		fprintf(out,"    Tcl_ResetResult(interp);");
	}

	fprintf(out,"	 if (objc < 3) {"
                "		 Tcl_WrongNumArgs(interp, 2, objv, \"variable\");"
				"        return TCL_ERROR;"
				"    }"
				"    if (Tcl_GetIndexFromObj(interp, objv[2],"
				"							 setgetPtr, \"variable\", 0, &__i) != TCL_OK) {"
				"    	return TCL_ERROR;"
				"    }"
				"    switch (__i) {");

	for(j=0; j<i; j++) {
		fprintf(out,"case %2d: ", j);
		OutputGet(j, 0, "__res = ", "");
		fprintf(out,"break;");
	}
	fprintf(out,"}\n%s\n", codeGet);
	fprintf(out,"Tcl_SetObjResult(interp, __res);"
				"return TCL_OK;\n}\n");


	fclose(in);
	fclose(out);
	return 0;
}
