#=================================================================================================
#
# The window in which to define the view/pose
#

proc library::pose:dialogSelect {entry mode} {
variable list
variable data

	# update selection
	foreach p $list($entry:part) {
		set joint [lindex $list($entry:joint) 0]
		if {$mode=="on"} {
			$data(v:body:internal) select $p
			$data(v:body:internal) set $joint [$data(v:body:model) get $joint]
		} elseif {$mode=="off"} {
			$data(v:body:internal) unselect $p
			$data(v:body:internal) set $joint [lindex $list($entry:joint) 1]
		}
	}
}


proc library::pose:dialogWindowInit {} {
global status
variable data
variable list

	set w [frame $data(w:pose:dialog).top]
	  # the view of the pose
	  set data(w:pose:view) [togl $w.t -width 75 -height 125 -double true \
								-ident "library" -depth true -class view3D -share $status(opengl)]
	    set v $data(w:pose:view)
		$v set axis false
		$v set floor false
		$v add $data(v:body:internal)
		replaceBindtags $v "OpenGM" "View3D"

	    menu $w.m -tearoff 0 -activeborderwidth 1 -bd 1 -font popupbutton
	      $w.m add check -label "Floor" -variable library::data(v:pose:floor) \
		      -command "$v set floor \[set library::data(v:pose:floor)\]"
	      $w.m add check -label "Axis" -variable library::data(v:pose:axis) \
		      -command "$v set axis \[set library::data(v:pose:axis)\]"
	      $w.m add check -label "Wireframe" -variable library::data(v:pose:wire) \
		      -command "$v set forcewire \[set library::data(v:pose:forcewire)\]"
	    bind $v <ButtonPress-3> "tk_popup $w.m %X %Y"
		
	  # the check list
	  set data(w:pose:list) [tixAutoCheckList $w.c -height 125]
		$w.c subwidget hlist configure -font checklist
		# the name of the part to select/unselect
		set list(0:part)	"Head"
		set list(1:part)	"body"
		set list(1.1:part)	"torso"
		set list(1.2:part)	"pelvis"
		set list(2:part)	"left:Arm"
		set list(2.1:part)	"left:arm"
		set list(2.2:part)	"left:forearm"
		set list(2.3:part)	"left:hand"
		set list(3:part)	"right:Arm"
		set list(3.1:part)	"right:arm"
		set list(3.2:part)	"right:forearm"
		set list(3.3:part)	"right:hand"
		set list(4:part)	"left:Leg"
		set list(4.1:part)	"left:thigh"
		set list(4.2:part)	"left:leg"
		set list(4.3:part)	"left:Foot"
		set list(5:part)	"right:Leg"
		set list(5.1:part)	"right:thigh"
		set list(5.2:part)	"right:leg"
		set list(5.3:part)	"right:Foot"

		# what to do with the joint
		set list(0:joint)	"head {0 0 0}"
		set list(1:joint)	"body {{0 0 0} {0 0 0}}"
		set list(1.1:joint)	"torso {0 0 0}"
		set list(1.2:joint)	"pelvis {0 0 0}"
		set list(2:joint)	"left:arm {{0 0 0 0} {0 0} {0 0 0}}"
		set list(2.1:joint)	"left:shoulder {0 0 0 0}"
		set list(2.2:joint)	"left:elbow {0 0}"
		set list(2.3:joint)	"left:wrist {0 0 0}"
		set list(3:joint)	"right:arm {{0 0 0 0} {0 0} {0 0 0}}"
		set list(3.1:joint)	"right:shoulder {0 0 0 0}"
		set list(3.2:joint)	"right:elbow {0 0}"
		set list(3.3:joint)	"right:wrist {0 0 0}"
		set list(4:joint)	"left:leg {{0 0 0} 0 {0 0 0} 0}"
		set list(4.1:joint)	"left:hip {0 0 0}"
		set list(4.2:joint)	"left:knee:X 0"
		set list(4.3:joint)	"left:foot {0 0 0 0}"
		set list(5:joint)	"right:leg {{0 0 0} 0 {0 0 0} 0}"
		set list(5.1:joint)	"right:hip {0 0 0}"
		set list(5.2:joint)	"right:knee:X 0"
		set list(5.3:joint)	"right:foot {0 0 0 0}"

		foreach {idx text} {0 "Head"
				   1 "Body"			1.1 "Torso" 1.2 "Pelvis"
				   2 "Left Arm"		2.1 "Left arm" 2.2 "Left forarm" 2.3 "Left hand"
				   3 "Right Arm"	3.1 "Right arm" 3.2 "Right forarm" 3.3 "Right hand"
				   4 "Left Leg"		4.1 "Left thigh" 4.2 "Left leg" 4.3 "Left foot"
				   5 "Right Leg"	5.1 "Right thigh" 5.2 "Right leg" 5.3 "Right foot"} {
			$w.c subwidget hlist add $idx -text $text
			$w.c setstatus $idx on
		}
		$w.c configure -callback library::pose:dialogSelect
		$w.c autosetmode

      pack $w.t -side left -padx 10 -pady 10
      pack $w.c -side left -expand yes -fill both -padx 10 -pady 10
	pack $w -expand yes -fill both

	# the name
	set data(w:pose:name) [tixLabelEntry $data(w:pose:dialog).name -label "Comment: "]
	  $data(w:pose:name) subwidget label configure -font entry
	  $data(w:pose:name) subwidget entry configure -font entry \
					-textvariable library::data(v:name)
	pack $data(w:pose:name) -expand yes -fill x -padx 5 -pady 5

	# the button box
	set w [tixButtonBox $data(w:pose:dialog).box -orientation horizontal]
	  $w add ok     -text OK -command "library::callback:ok $data(w:pose:dialog) pose" -width 5
	  $w add cancel -text Cancel -command "wm withdraw $data(w:pose:dialog)" -width 5
	pack $w -side bottom -fill x

	bind $data(w:pose:dialog) <Key-Return> "library::callback:ok $data(w:pose:dialog) pose"
}
# force init
library::pose:dialogWindowInit


#-------------------------------------------------------------------------------------------------
#	MODE add : 	modelView = path to 3d view
#				modelBody = body to use as a model
#				entry = entry idx in the popup menu -> category
#
#	MODE update :	modelView = parameter of 3d view
#
proc library::pose:dialog {mode modelView modelBody {entry {}}} {
variable list
variable data

	# check if a class if active
	if {$data(v:class) == ""} {
		tk_messageBox -icon "error" -title "Error" -type ok -default ok \
				-message "You must choose a class first"
		return
	}

	# copy the view point
	if {$mode == "add"} {
		$data(w:pose:view) set all [$modelView get all]
	} else {
		$data(w:pose:view) set all $modelView
	}
	$data(w:pose:view) set floor $data(v:pose:floor)
	$data(w:pose:view) set axis $data(v:pose:axis)
	$data(w:pose:view) set forcewire $data(v:pose:forcewire)

	# copy the pose
	set data(v:body:model) $modelBody
	$data(v:body:internal) set pose {{0.0 0.0 0.0} \
									 {{0.0 0.0 0.0} {0.0 0.0 0.0}} \
									 {{0.0 0.0 0.0 0.0} {0.0 0.0} {0.0 0.0 0.0}} \
									 {{0.0 0.0 0.0 0.0} {0.0 0.0} {0.0 0.0 0.0}} \
									 {{0.0 0.0 0.0} 0.0 {0.0 0.0 0.0} 0.0} \
									 {{0.0 0.0 0.0} 0.0 {0.0 0.0 0.0} 0.0} }
	$data(v:body:internal) unselect all
	foreach i {0 1 2 3 4 5} {
		$data(w:pose:list) setstatus $i off
		$data(w:pose:list) close $i
	}

	# update the selection
	if {$mode == "add"} {
		set data(v:tab) [$entry entrycget active -label]
		set entries $list(config:$data(v:tab))
	} else {
		set data(v:tab) $data(t:category)
		set entries $data(t:parts)
	}
	foreach i $entries {
		$data(w:pose:list) setstatus $i on
	}

	# set the name and configure the OK button
	focus $data(w:pose:name)
	if {$mode == "add"} {
		set data(v:name) ""
		set data(v:useID) -1
	} else {
		set data(v:name) $data($data(t:category):$data(t:id),name)
		set data(v:useID) $data(t:id)
	}

	# go for it
	wm deiconify $data(w:pose:dialog)
}


#=================================================================================================
#
# The window in which to define the view/position
#

proc library::position:dialogSelect {entry mode} {
variable list
variable data

	# update selection
	if {$mode=="on"} {
		if {$list($entry:joint) == "cog:rY"} {
			$data(v:body:internal) set cog:rY $data(v:cog:rY)
		} else {
			$data(v:body:internal) set $list($entry:joint) \
					[$data(v:body:model) get $list($entry:joint)]
		}
	} elseif {$mode=="off"} {
		$data(v:body:internal) set $list($entry:joint) 0
	}
}


proc library::position:dialogWindowInit {} {
global status
variable data
variable list


	set w [frame $data(w:position:dialog).top]
	  # the view of the pose
	  set data(w:position:view) [togl $w.t -width 75 -height 125 -double true \
								-ident "library" -depth true -class view3D -share $status(opengl)]
	    set v $data(w:position:view)
		$v add $data(v:body:internal)
		$v add $data(v:body:internal2)
		$v ortho
		$v set axis 0
		$v set view $data(v:position:top)
		replaceBindtags $v "OpengGL" "View3DProjection"

	    menu $w.m -tearoff 0 -font popupmenu -bd 1 -activeborderwidth 1 
	      $w.m add command -label "Toggle view (3D/Top)" -command "library::position:toggleView"
		  $w.m add separator
	      $w.m add check -label "Floor" -variable library::data(v:position:floor) \
		      -command "$v set floor \[set library::data(v:position:floor)\]"
	      $w.m add check -label "Axis" -variable library::data(v:position:axis) \
		      -command "$v set axis \[set library::data(v:position:axis)\]"
	      $w.m add check -label "Wireframe" -variable library::data(v:position:wire) \
		      -command "$v set forcewire \[set library::data(v:position:forcewire)\]"
	    bind $v <ButtonPress-3> "tk_popup $w.m %X %Y"
		
	  # the check list
	  set data(w:position:list) [tixAutoCheckList $w.c -height 125]
		$w.c subwidget hlist configure -font checklist
		# what to do with the joint
		set list(11:joint)	"cog:Y"
		set list(12:joint)	"cog:rY"
		set list(13:joint)	"cog:rX"
		set list(14:joint)	"cog:rZ"

		foreach {idx text} {11 "Altitude"
							12 "Orientation"
							13 "Roll (X)"
							14 "Bend (Z)" } {
			$w.c subwidget hlist add $idx -text $text
			$w.c setstatus $idx on
		}
		$w.c configure -callback library::position:dialogSelect
		$w.c autosetmode

	  # the name
	  set data(w:position:name) [tixLabelEntry $w.name -label "Name: "]
		$data(w:position:name) subwidget label configure -font entry
		$data(w:position:name) subwidget entry configure -font entry \
					-textvariable library::data(v:name)

	  # the reference body
	  menubutton $w.bodies -indicatoron 1 -menu $w.bodies.menu \
			  -font menubutton -textvariable library::data(v:position:bodyName) \
			  -relief raised -anchor w -direction below -bd 1
	  set data(m:bodies) [menu $w.bodies.menu -activeborderwidth 1 -bd 1 \
			  -font menubutton -tearoff 0]	  
	
	  # the layout
	  grid $w.bodies	x		-sticky news -padx 10 -pady 2
	  grid $w.t 		$w.c 	-sticky news -padx 10 -pady 2
	  grid $w.name		-		-sticky news -padx 10 -pady 10
	  grid rowconfig    $w {0 2} -weight 0
	  grid rowconfig    $w {1} -weight 1
	  grid columnconfig $w {0 1} -weight 1
	pack $w -side top -expand yes -fill both

	# the button box
	set w [tixButtonBox $data(w:position:dialog).box -orientation horizontal]
	  $w add ok -text OK -command "library::callback:ok $data(w:position:dialog) position" \
				-width 5
	  $w add cancel -text Cancel -command "wm withdraw $data(w:position:dialog)" -width 5
	pack $w -side bottom -fill x

	bind $data(w:position:dialog) <Key-Return> \
			"library::callback:ok $data(w:position:dialog) position"
}
# force init
library::position:dialogWindowInit


proc library::position:dialog {modelView body bodies} {
variable list
variable data


	# check if a class if active
	if {$data(v:class) == ""} {
		tk_messageBox -icon "error" -title "Error" -type ok -default ok \
				-message "You must choose a class first"
		return
	}

	# copy the view point
	if {$data(v:position:mode) == "3d"} {
		$data(w:position:view) set all [$modelView get all]
		$data(w:position:view) set floor $data(v:position:floor)
		$data(w:position:view) set axis $data(v:position:axis)
		$data(w:position:view) set forcewire $data(v:position:forcewire)
	}

	# build the list of possible references
	$data(m:bodies) delete 0 end
	catch {unset ref}
	foreach b $bodies {
		if {$b != $body} {
			$data(m:bodies) add command -label [$b get name] \
					-command "library::position:chooseBody $b"
			if {![info exists ref]} { set ref $b }
		}
	}

	# copy the pose : body = relative
	set data(v:body:model) $body
	wm deiconify $data(w:position:dialog)
	$data(v:body:internal) set all [$body get all]
	$data(v:body:internal) select all

	library::position:chooseBody $ref

	# update the selection
	set data(v:tab) Position
	foreach i $list(config:$data(v:tab)) {
		$data(w:position:list) setstatus $i on
	}

	# clear the name and give the focus
	focus $data(w:position:name)
	set data(v:name) ""
}

#-------------------------------------------------------------------------------------------------

proc library::position:toggleView {} {
variable data

	if {$data(v:position:mode) == "top"} {
		# top -> 3D
		set data(v:position:top) [$data(w:position:view) get view]
		$data(w:position:view) persp
		$data(w:position:view) set view $data(v:position:3d)
		set data(v:position:mode) 3d
		replaceBindtags $data(w:position:view) "View3DProjection" "View3D"

	} else {
		# 3D -> top
		set data(v:position:3d) [$data(w:position:view) get view]
		$data(w:position:view) ortho
		$data(w:position:view) set view $data(v:position:top)
		set data(v:position:mode) top
		replaceBindtags $data(w:position:view) "View3D" "View3DProjection"
	}
}

#-------------------------------------------------------------------------------------------------

proc library::position:chooseBody {body} {
variable data

	# get the pose for the reference
	$data(v:body:internal2) set pose [$body get pose]
	$data(v:body:internal2) set cog:rX [$body get cog:rX]
	$data(v:body:internal2) set cog:rZ [$body get cog:rZ]

	# update the option menu
	set data(v:position:bodyName) [$body get name]

	# update the model (with rotation of the position on the floor)
	set data(v:cog:rY) [expr [$data(v:body:model) get cog:rY]-[$body get cog:rY]]
	$data(v:body:internal) set cog:rY $data(v:cog:rY)
	set cos [expr cos(([$data(v:body:model) get cog:rY]-$data(v:cog:rY))*3.141592654/180)]
	set sin [expr sin(([$data(v:body:model) get cog:rY]-$data(v:cog:rY))*3.141592654/180)]

	set cogX [expr [$body get cog:X]-[$data(v:body:model) get cog:X]]
	set cogZ [expr [$body get cog:Z]-[$data(v:body:model) get cog:Z]]

	$data(v:body:internal) set cog:X [expr -$cos*$cogX - $sin*$cogZ]
	$data(v:body:internal) set cog:Z [expr  $sin*$cogX - $cos*$cogZ]

#puts "internal = [$data(v:body:internal) get cog]"
#puts "$cogX , $cogZ  ----  turn = [expr (-[$data(v:body:model) get cog:rY]+$data(v:cog:rY))]  ----    [expr $cos*$cogX + $sin*$cogZ] , [expr -$sin*$cogX + $cos*$cogZ]"
#puts "internal2 = [$data(v:body:internal2) get cog]"
}
