#include "togl.h"
#include <math.h>
#include <ctype.h>

#include "list.h"
#include "routine.h"
#include "track.h"
#include "movie.h"
#include "body.h"		// only for body->number in .i

extern char *trackName[];
extern char *jointName[];
extern Tcl_Interp *currentInterp;


//================================================================================================

void Stringify(FILE *out, char *str)
{
	fputc('"', out);

	while (*str) {
		switch (*str) {
			case '\n':	fputc('\\', out); fputc('n', out); break;
			case '\t':	fputc('\\', out); fputc('t', out); break;
			case '\r':	fputc('\\', out); fputc('r', out); break;
			case '\b':	fputc('\\', out); fputc('b', out); break;
			case '\f':	fputc('\\', out); fputc('f', out); break;
	
			default:	if (!isprint(*str))
							fprintf(out, "\0x%o", *str);
						else
							fputc(*str, out);

						break;
		}
		str++;
	}

	fputc('"', out);
}

//================================================================================================

int Routine::Update(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
static char *cmdPtr[] = { "cog", "joints", (char *) 0};
enum {cog, joints} cmd;
int dancer, start=0, end=0;
char *bodyName;
Tcl_CmdInfo info;


	if (objc < 4) {
		Tcl_WrongNumArgs(interp, 2, objv, "command body");
		return TCL_ERROR;
	}

	// cmd
	if (Tcl_GetIndexFromObj(interp, objv[2], cmdPtr, "command", 0, (int *) &cmd) != TCL_OK) {
		return TCL_ERROR;
	}
	// body
	bodyName = Tcl_GetStringFromObj(objv[3], NULL);
	if (Tcl_GetCommandInfo(interp, bodyName, &info) == 0 || *bodyName != 'b') {
		Tcl_AppendResult(interp, "no body named \"", bodyName, "\"",(char *) 0);
		return TCL_ERROR;
	}
	for(dancer = 0; dancer < nbDancers; dancer++) {
		if ((class Body *) info.objClientData == bodies[dancer])
			break;
	}
	if (dancer == nbDancers) {
		Tcl_AppendResult(interp, "body \"", bodyName, "\" not used in this routine",(char *) 0);
		return TCL_ERROR;
	}


	// special case of init pos
	if (time == 0 && cmd == cog) {
		int goon = FALSE;
		// reset dancer to zero
		if (dancers[dancer]->track[cogX].IsUsed()) {
			dancers[dancer]->track[cogX].JointValue();
		} else {
			goon = TRUE;
		}
		if (dancers[dancer]->track[cogZ].IsUsed()) {
			dancers[dancer]->track[cogZ].JointValue();
		} else {
			goon = TRUE;
		}
//fprintf(stderr, "\n%g %g",dancers[dancer]->track[cogX].data->first()->startValue,dancers[dancer]->track[cogZ].data->first()->startValue);
		dancerInitPos[dancer][4] = -dancerInitPos[dancer][4];
//fprintf(stderr, "=============== Routine::Update(1)  %d  --  %d:%d\n", dancer, 0, nbBeats);
		OrientDancer(dancer, 0, nbBeats);
		// record new init pos
		dancerInitPos[dancer][0] = bodies[dancer]->joints[cogX];
		dancerInitPos[dancer][1] = bodies[dancer]->joints[cogY];
		dancerInitPos[dancer][2] = bodies[dancer]->joints[cogZ];
		dancerInitPos[dancer][3] = bodies[dancer]->joints[cogRX];
		dancerInitPos[dancer][4] = bodies[dancer]->joints[cogRY];
		dancerInitPos[dancer][5] = bodies[dancer]->joints[cogRZ];
		// reorient dancer
//fprintf(stderr, "\n%g %g %g    %g %g %g",dancerInitPos[dancer][0],dancerInitPos[dancer][1],dancerInitPos[dancer][2],dancerInitPos[dancer][3],dancerInitPos[dancer][4],dancerInitPos[dancer][5]);

		if (dancers[dancer]->track[cogY].IsUsed()) {
			dancers[dancer]->track[cogY].JointValue();
		} else {
			goon = TRUE;
		}
		if (dancers[dancer]->track[cogRX].IsUsed()) {
			dancers[dancer]->track[cogRX].JointValue();
		} else {
			goon = TRUE;
		}
		if (dancers[dancer]->track[cogRZ].IsUsed()) {
			dancers[dancer]->track[cogRZ].JointValue();
		} else {
			goon = TRUE;
		}
//fprintf(stderr, "=============== Routine::Update(2)  %d  --  %d:%d\n", dancer, 0, nbBeats);
		OrientDancer(dancer, 0, nbBeats);

		if (!goon) {
			return TCL_OK;
		}
	}

	//------------------------------------------------------------------------------------
	// handle request - check if we need to force (recording of pos means cog:X AND cog:Z)
	int mode = NORMAL;
	enum {nothing, modifValue, newTrack, newKey};
	int type, type2;
	switch (cmd) {
		case cog:		start = cogX;
						end = cogRZ;
						type = dancers[dancer]->track[cogX].Update(time,
										bodies[dancer]->joints[cogX], nbBeats, QUERY);
						type2 = dancers[dancer]->track[cogZ].Update(time,
										bodies[dancer]->joints[cogZ], nbBeats, QUERY);
						if (type == newKey || type2 == newKey) {
							mode = FORCE;
						}
						break;
		case joints:	start = pelvisX;
						end = rFootX;
						break;
	}

	//-----------------------------
	// update the tracks in routine
	struct Tcl_Obj *res = Tcl_NewListObj(0, NULL);
	for(int i=start; i<=end; i++) {
//fprintf(stderr, "[%s / %s] @ %d -- new = %g\n", bodyName, trackName[i],time, bodies[dancer]->joints[i]);
		extern int trackToJoint[];
		// find the type of update needed
		type = dancers[dancer]->track[i].Update(time, bodies[dancer]->joints[i], nbBeats, QUERY);

		if (type == nothing) {
			//-------------------
			// nothing = continue
			continue;

		} else if (type == modifValue) {
			//--------------------------------------------------------------------
			// modifValue = modification of the routine and registration of modifs

			if (!(time==0 && cmd==cog)) {
				// check if new modif by calling registering with "false" modif
				class Modif m(VAL_VALUE | TEST_ONLY, time, bodies[dancer]->joints[i]);
				if (RegisterModif(dancer, i, &m)) {
					Tcl_ListObjAppendElement(interp, res,
							Tcl_NewStringObj(jointName[trackToJoint[i]], -1));
					Tcl_ListObjAppendElement(interp, res, Tcl_NewIntObj(time));
				}
			}
			if (time == nbBeats && dancers[dancer]->track[i].data->last()->isSynthetic == TRUE) {
				Tcl_ListObjAppendElement(interp, res,
						Tcl_NewStringObj(jointName[trackToJoint[i]], -1));
				Tcl_ListObjAppendElement(interp, res, Tcl_NewIntObj(time));				
			}
			// change the value --- no call to Track::Update
			ModifyValue(time, dancer, i, 0, -1, bodies[dancer]->joints[i]);

		} else {
			int lastTime = 0, tmpTime;
			//------------------------------------------------------
			// newTrack = search the last defined time for the joint
			if (type == newTrack) {
				extern int jointToTrack[][4];
				int joint = trackToJoint[i];
				for(int l=0; l<4 && jointToTrack[joint][l]!=-1; l++) {
					int track = jointToTrack[joint][l];
					tmpTime = dancers[dancer]->track[track].LastDefinedTime(time);
//fprintf(stderr, "%s : lastTime(%s,%d) = %d\n", trackName[i], trackName[track], time, tmpTime);
					if (tmpTime > lastTime) {
						lastTime = tmpTime;
					}
				}
			}
			// newTrack or newKey = create it
			type = dancers[dancer]->track[i].Update(time, bodies[dancer]->joints[i], nbBeats,
													i==cogX || i==cogZ ? mode : NORMAL, lastTime);
			if (time==0 && cmd==cog) {
				continue;
			}
			switch (type) {
				case newKey:	Tcl_ListObjAppendElement(interp, res,
										Tcl_NewStringObj(jointName[trackToJoint[i]], -1));
								Tcl_ListObjAppendElement(interp, res, Tcl_NewIntObj(time));
								break;
				case newTrack:	if (cmd != cog) {
									Tcl_ListObjAppendElement(interp, res,
											Tcl_NewStringObj(jointName[trackToJoint[i]], -1));
									Tcl_ListObjAppendElement(interp, res, Tcl_NewIntObj(0));
								}
								if (lastTime != 0) {
									Tcl_ListObjAppendElement(interp, res,
											Tcl_NewStringObj(jointName[trackToJoint[i]], -1));
									Tcl_ListObjAppendElement(interp, res,
											Tcl_NewIntObj(lastTime));
								}
								Tcl_ListObjAppendElement(interp, res,
										Tcl_NewStringObj(jointName[trackToJoint[i]], -1));
								Tcl_ListObjAppendElement(interp, res, Tcl_NewIntObj(time));
								break;
			}
			// try to register the modif in a figure as a new key
			class Modif m(VAL_VALUE | NEW_KEY, time, bodies[dancer]->joints[i]);
			if (type != modifValue && RegisterModif(dancer, i, &m)) {
				Tcl_ListObjAppendElement(interp, res,
						Tcl_NewStringObj(jointName[trackToJoint[i]], -1));
				Tcl_ListObjAppendElement(interp, res, Tcl_NewIntObj(time));
			}
		}
	}

	Tcl_SetObjResult(interp, res);
	return TCL_OK;
}

//------------------------------------------------------------------------------------------------

int Routine::Sync(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
int dancer = -1, time = -1, track = -1;

	if (objc < 5) {
		Tcl_WrongNumArgs(interp, 2, objv, "time dancer track");
		return TCL_ERROR;
	}

	// get the parameters
	Tcl_GetIntFromObj(interp, objv[2], &time);
	Tcl_GetIntFromObj(interp, objv[3], &dancer);
	if (Tcl_GetIndexFromObj(interp, objv[4], trackName, "track", 0, &track) != TCL_OK) {
		return TCL_ERROR;
	}
	if (dancer<0 || dancer>=nbDancers) {
		Tcl_SetResult(interp, "bad dancer index", 0);
		return TCL_ERROR;
	}

	dancers[dancer]->SetBody(bodies[dancer], time, track);

	return TCL_OK;
}

//================================================================================================

int Routine::HandleMovie(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
static char *cmdPtr[] = {"add", "delete", "changename", 0};
enum {add, del, change} index;
char *name, *tmp;
ListIterator<Movie> m;
List<Bezier3D> *eye, *center;


	if (objc < 4) {
		Tcl_WrongNumArgs(interp, 2, objv, "add/delete/changename name ?newname?");
		return TCL_ERROR;
	}

	// get the parameters
	if (Tcl_GetIndexFromObj(interp, objv[2], cmdPtr, "command", 0, (int *) &index) != TCL_OK) {
		return TCL_ERROR;
	}
	name = Tcl_GetStringFromObj(objv[3], NULL);

	// doit
	switch (index) {
		case add:	// check if name is unique
					for(m.reset(movies); m; ++m) {
						if (strcmp(m()->name(), name)==0) {
							Tcl_SetResult(interp, "duplicate name for movie", TCL_STATIC);
							return TCL_ERROR;
						}
					}

					// create an empty movie
					eye = new List<Bezier3D>;
					center = new List<Bezier3D>;
					Bezier3D *b;

					b = new Bezier3D(0, nbBeats);
					b->startX = b->startY = b->startZ = 0;
					b->leftX = b->leftY = b->leftZ = 0;
					b->rightX = b->rightY = b->rightZ = 0;
					b->endX = b->endY = b->endZ = 0;
					center->append(b);

					b = new Bezier3D(0, nbBeats);
					b->startX = 0; b->startY = 2; b->startZ = 7;
					b->leftX = 0; b->leftY = 2; b->leftZ = 7;
					b->rightX = 0; b->rightY = 2; b->rightZ = 7;
					b->endX = 0; b->endY = 2; b->endZ = 7;
					eye->append(b);

					// record the new movie
					tmp = new char[strlen(name)+1];
					strcpy(tmp, name);
					movies->append(new Movie(tmp,eye,center));
					break;

		case del:	// find it
					for(m.reset(movies); m; ++m) {
						if (strcmp(m()->name(), name) == 0) {
							movies->destroy(m());
							break;
						}
					}
					break;

		case change:// find it
					for(m.reset(movies); m; ++m) {
						if (strcmp(m()->name(), name) == 0) {
							if (objc < 5) {
								Tcl_WrongNumArgs(interp, 3, objv, "name newname");
								return TCL_ERROR;
							}
							tmp = Tcl_GetStringFromObj(objv[4], NULL);
							// check if name is unique
							for(ListIterator<Movie> mm(movies); mm; ++mm) {
								if (strcmp(mm()->name(), tmp)==0) {
									Tcl_SetObjResult(interp, Tcl_NewBooleanObj(TRUE));
									return TCL_OK;
								}
							}

							m()->NewName(tmp);
							Tcl_SetObjResult(interp, Tcl_NewBooleanObj(FALSE));
							break;
						}
					}
					break;
	}

	return TCL_OK;
}

//------------------------------------------------------------------------------------------------

int Routine::HandleDancer(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
static char *cmdPtr[] = {"add", "delete", 0};
enum {add, del} index;
char *name;


	if (objc < 4) {
		Tcl_WrongNumArgs(interp, 2, objv, "add/delete name");
		return TCL_ERROR;
	}

	// get the parameters
	if (Tcl_GetIndexFromObj(interp, objv[2], cmdPtr, "command", 0, (int *) &index) != TCL_OK) {
		return TCL_ERROR;
	}
	name = Tcl_GetStringFromObj(objv[3], NULL);

	// doit
	switch (index) {
		case add:	if (nbDancers >= MAX_DANCERS) {
						Tcl_AppendResult(interp, "too many dancers", (char *) 0);
						return TCL_ERROR;
					}
					dancerNames[nbDancers] = new char[strlen(name)+1];
					strcpy(dancerNames[nbDancers], name);
					dancers[nbDancers] = new Interpolator;
					nbDancers++;
					break;

		case del:	
					break;
	}

	return TCL_OK;
}

//================================================================================================

int Routine::Insert(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
int timeInsert, nb, asKeys, applyOnly, i;
char *str;
struct Tcl_Obj **list;
struct Tcl_CmdInfo info;
class Routine *source;
class Figure *figure;
Tcl_Obj *res = 0, *ids, *keysList;


	if (objc < 10) {
		Tcl_WrongNumArgs(interp, 2, objv, "routine filename color includeNotes asKeys"
										  " applyOnly time {dancer1 dancer2 ...}"
										  " ?{{reject list dancer 1} {reject list dancer 2}}?");
		return TCL_ERROR;
	}

	figure = new Figure();

	// routine name
	str = Tcl_GetStringFromObj(objv[2], NULL);
	if (Tcl_GetCommandInfo(interp, str, &info) == 0 || *str != 'r') {
		Tcl_AppendResult(interp, "no routine named \"",str,"\"",(char *) NULL);
		return TCL_ERROR;
	}
	source = (class Routine *) info.objClientData;
//	for(i=0; i<source->nbDancers; i++) {
//		source->dancers[i]->Smooth(0, source->nbBeats, FALSE);
//	}

	// filename
	str = Tcl_GetStringFromObj(objv[3], NULL);
	figure->filename = new char[strlen(str)+1];
	strcpy(figure->filename, str);

	// color
	str = Tcl_GetStringFromObj(objv[4], NULL);
	figure->color = new char[strlen(str)+1];
	strcpy(figure->color, str);

	// some bools parameters
	Tcl_GetBooleanFromObj(interp, objv[5], &figure->includeNotes);
	Tcl_GetBooleanFromObj(interp, objv[6], &asKeys);
	Tcl_GetBooleanFromObj(interp, objv[7], &applyOnly);

	// get the time
	Tcl_GetIntFromObj(interp, objv[8], &timeInsert);
	figure->start = timeInsert;
	if (figure->start > nbBeats)
		figure->start = nbBeats;
	figure->end = timeInsert+source->nbBeats;

	// get the map
	Tcl_ListObjGetElements(interp, objv[9], &nb, &list);
	if (nb != source->nbDancers) {
		Tcl_AppendResult(interp, "bad map (not matching number of dancer in figure)",
						 (char *) NULL);
		return TCL_ERROR;
	}
	figure->nbMap = nb;
	for(i=0; i<nb; i++) {
		Tcl_GetIntFromObj(NULL, list[i], &figure->map[i]);
	}

	// get the reject list
	if (objc == 11) {
		Tcl_ListObjGetElements(interp, objv[10], &nb, &list);
		if (nb != source->nbDancers) {
			Tcl_AppendResult(interp, "bad rejection list (not matching number"
							 " of dancer in figure)", (char *) NULL);
			return TCL_ERROR;
		}
		for(i=0; i<figure->nbMap; i++) {
			extern int jointToTrack[][4];
			Tcl_Obj **reject, *element;
			Tcl_ListObjGetElements(interp, list[i], &nb, &reject);
			for(int j=0; j<nb; j++) {
				int joint;
				int result = Tcl_GetIndexFromObj(interp, reject[j], jointName,
												 "joint", 0, &joint);
				if (result != TCL_OK || result >= NB_JOINTS) {
					Tcl_AppendResult(interp, "bad rejection joint", (char *) NULL);
					return TCL_ERROR;
				} else {
					for(int t=0; t<4 && jointToTrack[joint][t]!=-1; t++)
						figure->flags[i][jointToTrack[joint][t]] |= REJECT;
				}
			}
		}
	}

	//-----------------
	// do the insertion

	// 1. shift the routine to free some space & update nb beats in routine
	// 1.1. build the list of dependencies among dancers
	int doDancer[MAX_DANCERS], doShift = TRUE;
	for(i=0; i<MAX_DANCERS; i++) {
		doDancer[i] = FALSE;
	}
	for(i=0; i<figure->nbMap; i++) {
		if (figure->map[i] == -1) {
			continue; 		// if dancer not used in the mapping
		}
		doDancer[figure->map[i]] = TRUE;
		if (applyOnly) {
			continue;
		}
		// find dependencies
		for(ListIterator<Figure> f(figures); f; ++f) {
			if (f()->start >= timeInsert && f()->ApplyToDancer(figure->map[i])) {
				for(int j=0; j<f()->nbMap; j++) {
					if (f()->map[j] != -1)
						doDancer[f()->map[j]] = TRUE;
				}
				// shift figure to its new start/end (do it once only)
				if (doShift) {
					f()->start += source->nbBeats;
					f()->end += source->nbBeats;
				}
			}
		}
		doShift = FALSE;
	}
	// 1.2. for all the dancers linked together, do the shift (if needed)
	//      and build list of modified dancers
	ListIterator<Note> n;
	ids = Tcl_NewListObj(0, NULL);
	for (i=0; i<nbDancers; i++) {
		if (doDancer[i]) {
			if (!applyOnly) {
				// 1.2.1. keys
				dancers[i]->TranslateTime(source->nbBeats, timeInsert);
				// 1.2.2. notes
				for(n.reset(notes[i][REAL]); n; ++n) {
					if (n()->start() >= timeInsert)
						n()->offset(source->nbBeats);
				}
				for(n.reset(notes[i][INCLUDED]); n; ++n) {
					if (n()->start() >= timeInsert)
						n()->offset(source->nbBeats);
				}
			}
			Tcl_ListObjAppendElement(interp, ids, Tcl_NewIntObj(i));
		}
	}
	if (!applyOnly) {
		for(n.reset(notes[COMMENTS][REAL]); n; ++n) {
			if (n()->start() >= timeInsert)
				n()->offset(source->nbBeats);
		}
		// add -1 to force the move of notes
		Tcl_ListObjAppendElement(interp, ids, Tcl_NewIntObj(-1));
	}
	// 1.3. update nb beats of main routine
	if (!applyOnly) {
		nbBeats += source->nbBeats;
	}

	// 2. insert the figure
	if (!asKeys) {
		if (figures == 0) {
			figures = new List<Figure>;
		}
		figures->append(figure);
	}
	if (!figure->Insert(*this, *source)) {
		return TCL_ERROR;
	}

	// 3. final smooth of the routine
	for(i=0; i<nbDancers; i++) {
		if (doDancer[i]) {
			dancers[i]->Smooth(0, nbBeats, TRUE);
		}
	}

	// 4. generate result
	res = Tcl_NewListObj(0, NULL);
	Tcl_ListObjAppendElement(interp, res, Tcl_NewIntObj(asKeys ? -1 : figure->number));
	Tcl_ListObjAppendElement(interp, res, ids);

	keysList = Tcl_NewListObj(0, NULL);
	for(int d=0; d<figure->nbMap; d++) {
		extern int trackToJoint[];
		int alreadyDone[NB_JOINTS];
		for(i=0; i<NB_JOINTS; i++)
			alreadyDone[i] = FALSE;

		ids = Tcl_NewListObj(0, NULL);
		if (figure->map[d] != -1) {
			for(i=0; i<NB_DOF; i++) {
				if (!alreadyDone[trackToJoint[i]]
						&& (figure->flags[d][i] & USED)
						&& !(figure->flags[d][i] & REJECT)) {
					Tcl_ListObjAppendElement(interp, ids,
							Tcl_NewStringObj(jointName[trackToJoint[i]], -1));
					alreadyDone[trackToJoint[i]] = TRUE; 
				}
			}
		}
		Tcl_ListObjAppendElement(interp, keysList, ids);
	}
	Tcl_ListObjAppendElement(interp, res, keysList);

	// 5. cleanup if needed
	if (asKeys) {
		delete figure;
	}

	Tcl_SetObjResult(interp, res);
	return TCL_OK;
}

//------------------------------------------------------------------------------------------------

int Routine::CheckInsertion(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
int nb, timeInsert, dancer, applyOnly;
struct Tcl_Obj **list, *conflict[MAX_DANCERS], *res;
struct Tcl_CmdInfo info;
class Routine *source;
int map[MAX_DANCERS], i;
char *str;


	if (objc != 6) {
		Tcl_WrongNumArgs(interp, 2, objv, "routine applyOnly time {dancer1 dancer2 ...}");
		return TCL_ERROR;
	}

	// routine name
	str = Tcl_GetStringFromObj(objv[2], NULL);
	if (Tcl_GetCommandInfo(interp, str, &info) == 0 || *str != 'r') {
		Tcl_AppendResult(interp, "no routine named \"",str,"\"",(char *) NULL);
		return TCL_ERROR;
	}
	source = (class Routine *) info.objClientData;

	// get if apply only
	Tcl_GetBooleanFromObj(interp, objv[3], &applyOnly);

	// get the time
	Tcl_GetIntFromObj(interp, objv[4], &timeInsert);
	if (timeInsert > nbBeats) {
		Tcl_AppendResult(interp, "bad time (exceeding beats in routine",(char *) NULL);
		return TCL_ERROR;
	}

	// get the map
	Tcl_ListObjGetElements(interp, objv[5], &nb, &list);
	if (nb != source->nbDancers) {
		Tcl_AppendResult(interp, "bad map (not matching number of dancer in figure)",
						 (char *) NULL);
		return TCL_ERROR;
	}
	for(i=0; i<nb; i++) {
		Tcl_GetIntFromObj(NULL, list[i], &map[i]);
	}

	// check the conflicting tracks if we try to apply the figure
	res = Tcl_NewListObj(0, NULL);
	for(i=0; i<source->nbDancers; i++) {
		conflict[i] = Tcl_NewListObj(0, NULL);
		Tcl_ListObjAppendElement(interp, res, conflict[i]);
	}
	for(i=0; i<nb && applyOnly; i++) {
		if (map[i] == -1) {
			continue; 		// if dancer not used in the mapping
		}
		dancer = map[i];

		// check the joints
		int min, max;
		for (int j=0; j<NB_JOINTS; j++) {
			if (SearchConflict(*source, &min, &max, i, dancer, j, timeInsert)) {
				Tcl_ListObjAppendElement(interp, conflict[dancer],
						Tcl_NewStringObj(jointName[j], -1));
			}
		}
	}

	Tcl_SetObjResult(interp, res);
	return TCL_OK;
}

//================================================================================================

int Routine::GetFigureInfo(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) const
{
static char *varPtr[] = {
	"color", "filename", "name", "includenotes",
	"start", "end", "map", "jointsusedby", 
	"reject", 0};
enum {
	color, filename, name, notes,
	start, end, map, usedby,
	reject } index;
int number, dancer, dancerMap, joint, i;
Tcl_Obj *res = 0;


	if (objc < 4) {
		Tcl_WrongNumArgs(interp, 2, objv, "figureID var ?value (color/notes only)?");
		return TCL_ERROR;
	}

	// get the parameters
	Tcl_GetIntFromObj(interp, objv[2], &number);	
	if (Tcl_GetIndexFromObj(interp, objv[3], varPtr, "variable", 0, (int *) &index) != TCL_OK) {
		return TCL_ERROR;
	}

	// get var
	ListIterator<Figure> f;
	for (f.reset(figures); f; ++f) {
		if (f()->number == number) {
			switch (index) {
				case color:		if (objc == 5) {
									char *color = Tcl_GetStringFromObj(objv[4], NULL);
									delete[] f()->color;
									f()->color = new char[strlen(color)];
									strcpy(f()->color, color);
								}
								res = Tcl_NewStringObj(f()->color, -1);
								break;
				case filename:	res = Tcl_NewStringObj(f()->filename, -1); break;
				case name:		res = Tcl_NewStringObj(f()->name, -1); break;
				case start:		res = Tcl_NewIntObj(f()->start); break;
				case end:		res = Tcl_NewIntObj(f()->end); break;
				case notes:		if (objc == 5) {
									Tcl_GetBooleanFromObj(interp, objv[4], &f()->includeNotes);
								}
								res = Tcl_NewBooleanObj(f()->includeNotes);
								break;
				case map:
					res = Tcl_NewListObj(0, NULL);
					for(i=0; i<f()->nbMap; i++) {
						Tcl_ListObjAppendElement(interp, res, Tcl_NewIntObj(f()->map[i]));
					}
					break;

				case usedby:
					if (objc < 5) {
						Tcl_WrongNumArgs(interp, 4, objv, "dancer");
						return TCL_ERROR;
					}
					Tcl_GetIntFromObj(interp, objv[4], &dancer);
					if (dancer<0 || dancer>=f()->nbMap) {
						Tcl_SetResult(interp, "bad dancer map", 0);
						return TCL_ERROR;
					}
					res = Tcl_NewListObj(0, NULL);

					extern int trackToJoint[];
					int alreadyDone[NB_JOINTS];
					for(i=0; i<NB_JOINTS; i++)
						alreadyDone[i] = FALSE;

					for(i=0; i<NB_DOF; i++) {
						if (!alreadyDone[trackToJoint[i]]
								&& (f()->flags[dancer][i] & USED)
								&& !(f()->flags[dancer][i] & REJECT)) {
							Tcl_ListObjAppendElement(interp, res,
									Tcl_NewStringObj(jointName[trackToJoint[i]], -1));
							alreadyDone[trackToJoint[i]] = TRUE; 
						}
					}
					break;

				case reject:
					if (objc < 6) {
						Tcl_WrongNumArgs(interp, 4, objv, "dancer joint");
						return TCL_ERROR;
					}
					// get dancer
					Tcl_GetIntFromObj(interp, objv[4], &dancer);
					if (dancer<0 || dancer>=nbDancers) {
						Tcl_SetResult(interp, "bad dancer index", 0);
						return TCL_ERROR;
					}
					dancerMap = -1;
					for(i=0; i<f()->nbMap; i++) {
						if (f()->map[i] == dancer) {
							dancerMap = i;
							break;
						}
					}
					if (dancerMap == -1) {
						Tcl_SetResult(interp, "dancer is not mapped in figure", 0);
						return TCL_ERROR;
					}
					// analyze the joint
					if (Tcl_GetIndexFromObj(interp, objv[5], jointName, "joint",
											0, &joint) != TCL_OK) {
						return TCL_ERROR;
					}
					if (joint >= NB_JOINTS) {
						Tcl_SetResult(interp, "bad joint", 0);
						return TCL_ERROR;
					}
					// delete the point
					extern int jointToTrack[][4];
					for(i=0; i<4 && jointToTrack[joint][i]!=-1; i++) {
						dancers[dancer]->track[jointToTrack[joint][i]].Delete(f()->start+1,
								f()->end-1, nbBeats);
					}
					// do the rejection
					for(i=0; i<4 && jointToTrack[joint][i]!=-1; i++) {
						f()->flags[dancerMap][jointToTrack[joint][i]] &= ~USED;
						f()->flags[dancerMap][jointToTrack[joint][i]] |= REJECT;
					}

					res = Tcl_NewBooleanObj(TRUE);
					break;
			}
			break;
		}
	}

	if (!f) {
		Tcl_AppendResult(interp, "bad routine id", (char *) 0);
		return TCL_ERROR;
	}

	Tcl_SetObjResult(interp, res);
	return TCL_OK;
}

//------------------------------------------------------------------------------------------------

int Routine::HandleNote(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
static char *cmdPtr[] = {
	"add", "delete", "get", "exists", "modify", "next", "prev", "text", 0};
enum {add, del, get, exists, modify, next, prev, txt} index;

int dancer = -1, time = -1, timeEnd = -1, arg, id = -1, min = 0, max = nbBeats, i, t;
char *text = 0, *tmp;
ListIterator<Note> n;
Note *note;
Tcl_Obj *res = 0;


	if (objc < 4) {
		Tcl_WrongNumArgs(interp, 2, objv, "command args");
		return TCL_ERROR;
	}

	// get the parameters
	if (Tcl_GetIndexFromObj(interp, objv[2], cmdPtr, "command", 0, (int *) &index) != TCL_OK) {
		return TCL_ERROR;
	}

	switch (index) {
		case add:	if (objc < 6) {
						Tcl_WrongNumArgs(interp, 3, objv, "time ?timeEnd? dancer text");
						return TCL_ERROR;
					}
					arg = 3;
					// time(s)
					Tcl_GetIntFromObj(interp, objv[arg++], &time);
					timeEnd = time;
					if (objc > 6) {
						Tcl_GetIntFromObj(interp, objv[arg++], &timeEnd);
					}
					// dancer
					Tcl_GetIntFromObj(interp, objv[arg++], &dancer);
					if (dancer == -1) {
						dancer = MAX_DANCERS;
					} else if (dancer<0 || dancer>=nbDancers) {
						Tcl_SetResult(interp, "bad dancer index", 0);
						return TCL_ERROR;
					}
					// generate new note
					text = Tcl_GetStringFromObj(objv[arg++], NULL);
					tmp = new char[strlen(text)+1];
					strcpy(tmp, text);
					note = new Note(time, timeEnd, tmp);
					if (notes[dancer][REAL] == 0) {
						notes[dancer][REAL] = new List<Note>;
					}
					notes[dancer][REAL]->append(note);

					res = Tcl_NewIntObj(note->number());
//fprintf(stderr, "<<add = %d>>\n", note->number());
					break;

		case exists:
		case txt:	if (objc != 5) {
						Tcl_WrongNumArgs(interp, 3, objv, "time dancer");
						return TCL_ERROR;
					}
					Tcl_GetIntFromObj(interp, objv[3], &time);
					Tcl_GetIntFromObj(interp, objv[4], &dancer);
					if (dancer == -1) {
						dancer = MAX_DANCERS;
					} else if (dancer<0 || dancer>=nbDancers) {
						Tcl_SetResult(interp, "bad dancer index", 0);
						return TCL_ERROR;
					}

					n.reset(notes[dancer][REAL]);
					while (n) {
						if ((n()->start() < time && time <= n()->end())
						   || (time == 0)) {
							break;	// found
						}
						++n;
					}

					if (index == exists) {
						res = Tcl_NewIntObj(n ? n()->end()-n()->start() : -1);
//fprintf(stderr, "<<exists = %d>>\n", n ? n()->end()-n()->start() : -1);
					} else {
						if (!n) {
							n.reset(notes[dancer][INCLUDED]);
							while (n) {
								if ((n()->start() < time && time <= n()->end())
								   || (time == 0)) {
									break;	// found
								}
								++n;
							}
						}
						res = Tcl_NewStringObj((char *) (n ? n()->text() : ""), -1);
//fprintf(stderr, "<<text = %d - %d>>\n", text);
					}
					break;

		case del:
		case get:	if (objc != 4) {
						Tcl_WrongNumArgs(interp, 3, objv, "id");
						return TCL_ERROR;
					}
					Tcl_GetIntFromObj(interp, objv[3], &id);

					// search note & get the text
					for(i=0; !n && i<MAX_DANCERS+1; i++) {
						for(t=0; !n && t<2; t++) {
							n.reset(notes[i][t]);
							while (n) {
								if (n()->number() == id) {
									goto found;	// found
								}
								++n;
							}
						}
					}
					found:
					if (index == del) {
//fprintf(stderr, "<<del>>\n");
						res = Tcl_NewIntObj(n ? n()->number() : -1);
						if (i<MAX_DANCERS+1 && n) notes[i][REAL]->destroy(n());
					} else {
//fprintf(stderr, "<<get>>\n");
						res = Tcl_NewStringObj((char *) (n ? n()->text() : ""), -1);
					}
					break;

		case next:
		case prev:	if (objc != 5) {
						Tcl_WrongNumArgs(interp, 3, objv, "time dancer");
						return TCL_ERROR;
					}
					Tcl_GetIntFromObj(interp, objv[3], &time);
					Tcl_GetIntFromObj(interp, objv[4], &dancer);
					if (dancer == -1) {
						dancer = MAX_DANCERS;
					} else if (dancer<0 || dancer>=nbDancers) {
						Tcl_SetResult(interp, "bad dancer index", 0);
						return TCL_ERROR;
					}

					n.reset(notes[dancer][REAL]);
					while (n) {
						if (n()->start() < time && n()->end() >= min) {
							min = n()->end();
						}
						if (n()->start() >= time && n()->start() <= max) {
							max = n()->start();
						}
						++n;
					}

					res = Tcl_NewIntObj(index == prev ? min : max);
//fprintf(stderr, "<<prev/next = %d - %d>>\n", min,max);
					break;

		case modify:if (objc < 5) {
						Tcl_WrongNumArgs(interp, 3, objv, "noteid start/-1 ?end/-1? ?text?");
						return TCL_ERROR;
					}
					// note unique id
					Tcl_GetIntFromObj(interp, objv[3], &id);

					// time(s)
					Tcl_GetIntFromObj(interp, objv[4], &time);
					if (objc > 5) {
						Tcl_GetIntFromObj(interp, objv[5], &timeEnd);
					}
					// text
					if (objc > 6) {
						text = Tcl_GetStringFromObj(objv[6], NULL);
					}

					// search note & do modification
					for(i=0; !n && i<MAX_DANCERS+1; i++) {
						n.reset(notes[i][REAL]);
						while (n) {
							if (n()->number() == id) {
								break;	// found
							}
							++n;
						}
					}
					if (n) {
						n()->modify(time, timeEnd, text);
					}
					res = Tcl_NewIntObj(n);
//fprintf(stderr, "<<modify  %d-%d  %s>>\n",time,timeEnd,text);
					break;
	}

	Tcl_SetObjResult(interp, res);
	return TCL_OK;
}

//================================================================================================

int Routine::Load(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
int mode = COMPILE;
char *filename;


	if (objc < 3) {
		Tcl_WrongNumArgs(interp, 2, objv, "filename");
		return TCL_ERROR;
	}
	filename = Tcl_GetStringFromObj(objv[2], 0);

	// some settings
	extern Tcl_Interp *currentInterp;
	currentInterp = interp;

	// compile
	Tcl_ResetResult(interp);
	if (strcmp(Tcl_GetStringFromObj(objv[1], 0), "parse") == 0) {
		mode = PARSE;
	}

	if (Compile(filename, mode) == FALSE) {
		return TCL_ERROR;
	}

	return TCL_OK;
}

//------------------------------------------------------------------------------------------------

int Routine::Save(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
char *filename;
FILE *out;
int i, j, k, l;


	if (objc < 3) {
		Tcl_WrongNumArgs(interp, 2, objv, "filename");
		return TCL_ERROR;
	}
	filename = Tcl_GetStringFromObj(objv[2], 0);

	// save the file
	if ((out = fopen(filename, "w")) == 0) {
//		Tcl_SetResult(interp, "can't open file to save", TCL_STATIC);
//		return TCL_ERROR;
out = stderr;
	}

	//---------------
	// 1. global info
	fprintf(out, "#---------------------------------------\n");
	fprintf(out, "# Description\n#---------------------------------------\n\n");
	fprintf(out, "beats = %d\n", nbBeats);
	fprintf(out, "rhythm = %d\n", rhythm);
	fprintf(out, "bpm = %d\n\n", bpm);
	fprintf(out, "dancers = ");
	for(i=0; i<nbDancers; i++) {
		fprintf(out, "{%s, ", dancerSex[i] == MAN ? "man" : "woman");
		Stringify(out, dancerNames[i]);
		if (dancerTypes[i]) {
			fprintf(out, ", ");
			Stringify(out, dancerTypes[i]);
		}
		fprintf(out, "}");
		fprintf(out, i==nbDancers-1 ? "\n\n" : ", ");
	}

	if (name) {
		fprintf(out, "name = ");
		Stringify(out, name);
		fprintf(out, "\n");
	}
	if (description) {
		fprintf(out, "description = ");
		Stringify(out, description);
		fprintf(out, "\n");
	}
	if (copyright) {
		fprintf(out, "copyright = ");
		Stringify(out, copyright);
		fprintf(out, "\n");
	}
	if (text) {
		fprintf(out, "text = ");
		Stringify(out, text);
		fprintf(out, "\n");
	}
	if (name || description || copyright || text) fprintf(out, "\n");

	//--------
	// 2. keys
	fprintf(out, "\n#---------------------------------------\n");
	fprintf(out, "# Keys\n#---------------------------------------\n\n");

	for(i=0; i<nbDancers; i++) {
		// 2.1. init statement
		double value = 0;
		fprintf(out, "init(%d) = {\n", i);
		for(j=cogX; j<=cogRZ; j++) {
			if (dancers[i]->track[j].IsUsed()
				&& fabs(value = dancers[i]->track[j][0]) > .00001) {
				fprintf(out, "   %s = %g\n", trackName[j], value);
			}
		}
		fprintf(out, "}\n");

		// 2.2. keys defined in routine
		int mainHeader = FALSE, endTime = 0, saveGroup;
		double endValue = 0;
		List<TimeRange> *times;

		for(int group=0; group<NB_JOINTS; group++) {
			// skip second part of head
			if (group == 3) {
				continue;
			}

			// if one track of a 'joint' group is used, save all
			extern int jointToTrack[][4];
			saveGroup = FALSE;
			for(j=0; j<4 && jointToTrack[group][j]!=-1; j++) {
				if (dancers[i]->track[jointToTrack[group][j]].IsUsed()) {
					saveGroup = TRUE;
				}
			}

			// do it
			for(l=0; l<4 && jointToTrack[group][l]!=-1; l++) {
				j = jointToTrack[group][l];
				if (dancers[i]->track[j].IsUsed()) {
					int trackHeader = FALSE, firstRange = TRUE, finished = FALSE;
					// 2.2.1. key the key time
					times = ListKeysTime(i, j);

					ListIterator<TimeRange> t(times);
					ListIterator<Chunk> c(dancers[i]->track[j].data);

//fprintf(stderr, "    %-20s = ", trackName[j]);for(ListIterator<TimeRange> _t(times); _t; ++_t) {fprintf(stderr, "    %d-%d  ", _t()->start, _t()->end);}fprintf(stderr, "\n");

					// 2.2.2. write the key if not in figure & not synthetic
					while(t) {
						// find start of time range in chunk
						while (c && c()->startTime < t()->start)
							++c;

						if (!c)
							break; // no data for the track in the time range

					// while in time range
//fprintf(stderr, "dancer = %d -- track = %s -- c = %d:%d\n", i, trackName[j], c()->startTime, c()->endTime);
						while (c && c()->startTime < t()->end && c()->endTime <= t()->end) {
							if (!c()->isSynthetic) {
								// print the needed headers
								if (!mainHeader) {
									mainHeader = TRUE;
									fprintf(out, "key(%d) @ [%d:%d] = {\n", i, 0, nbBeats);
								}
								if (!trackHeader) {
									trackHeader = TRUE;
									fprintf(out, "    %s =", trackName[j]);
								}
								if (!firstRange) {
									fprintf(out, ",");
									firstRange = TRUE;
								}

								c()->Save(out);
								finished = FALSE;
								endTime = c()->endTime;
								endValue = c()->endValue;
							} else if (c()->startTime == 0) {
								// print the needed headers
								if (!mainHeader) {
									mainHeader = TRUE;
									fprintf(out, "key(%d) @ [%d:%d] = {\n", i, 0, nbBeats);
								}
								if (!trackHeader) {
									trackHeader = TRUE;
									fprintf(out, "    %s =", trackName[j]);
								}
								if (!firstRange) {
									fprintf(out, ",");
								} else {
									firstRange = FALSE;
								}
								// start value at 0
								finished = FALSE;
								endTime = c()->startTime;
								endValue = c()->startValue;
							} else {
								if (trackHeader && !finished) {
									finished = TRUE;
									fprintf(out, " %g@%d", endValue, endTime);
									firstRange = FALSE;
								}
							}
							++c;
						}
						if (trackHeader && !finished) {
							fprintf(out, " %g@%d", endValue, endTime);
							firstRange = FALSE;
						}

						// advance to next time range
						++t;
					}
					if (trackHeader) {
						fprintf(out, "\n");
					}

					// 2.2.3. free local variables
					times->free();
					delete times;

				} else if (saveGroup) {
					// print the needed headers
					if (!mainHeader) {
						mainHeader = TRUE;
						fprintf(out, "key(%d) @ [%d:%d] = {\n", i, 0, nbBeats);
					}
					fprintf(out, "    %s = 0@0\n", trackName[j]);
				}
			}
		}
		if (mainHeader) fprintf(out, "}\n");
		fprintf(out, "\n");
	}

	//-----------
	// 3. figures
	fprintf(out, "\n#---------------------------------------\n");
	fprintf(out, "# Figures\n#---------------------------------------\n\n");
	for (ListIterator<Figure> f(figures); f; ++f) {
		// save the figure
		f()->Save(out);
	}

	//---------
	// 4. notes
	fprintf(out, "\n#---------------------------------------\n");
	fprintf(out, "# Notes\n#---------------------------------------\n\n");
	for(i=0; i<MAX_DANCERS+1; i++) {
		ListIterator<Note> n(notes[i][REAL]);
		if (n) {
			fprintf(out, "notes(%d) = ", i == MAX_DANCERS ? -1 : i);
			for(; n; ++n) {
				if (!n.isHead()) fprintf(out, ",\n           ");

				Stringify(out, n()->text());
				if (n()->start() != n()->end())
					fprintf(out, "@[%d:%d]", n()->start(), n()->end());
				else
					fprintf(out, "@%d", n()->start());
			}
			fprintf(out, "\n");
		}
	}

	//----------
	// 5. movies
	fprintf(out, "\n#---------------------------------------\n");
	fprintf(out, "# Movies\n#---------------------------------------\n\n");
	for(ListIterator<Movie> m(movies); m; ++m) {
		m()->Save(out);
	}


if (out!=stderr)
	fclose(out);
	return TCL_OK;
}

//================================================================================================

int Routine::ApplyTo(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) const
{
int dancer = -1, time = -1;
class Body *body;
char *str;
struct Tcl_CmdInfo info;


	if (objc != 5) {
		Tcl_WrongNumArgs(interp, 2, objv, "dancer time body");
		return TCL_ERROR;
	}

	// get dancer index
	Tcl_GetIntFromObj(NULL, objv[2], &dancer);
	if (dancer<0 || dancer>MAX_DANCERS-1) {
		Tcl_AppendResult(interp,"bad index", (char *) NULL);
		return TCL_ERROR;
	}

	// get time
	Tcl_GetIntFromObj(NULL, objv[3], &time);
	if (time<0 || time>nbBeats) {
		Tcl_AppendResult(interp,"bad time", (char *) NULL);
		return TCL_ERROR;
	}

	// get body
	str = Tcl_GetStringFromObj(objv[4], NULL);
	if (Tcl_GetCommandInfo(interp, str, &info) == 0 || *str != 'b') {
		Tcl_AppendResult(interp, "no body named \"", str, "\"", (char *) NULL);
		return TCL_ERROR;
	}
	body = (class Body *) info.objClientData;

	// do it
	dancers[dancer]->SetBody(body, time);

	return TCL_OK;
}


//================================================================================================

#include "routine.icc"

//------------------------------------------------------------------------------------------------

static int Routine_ObjCmd(ClientData clientData, Tcl_Interp *interp,
					   int objc, Tcl_Obj *CONST objv[])
{
struct Routine *routine = (struct Routine *) clientData;
static char *optionsPtr[] = {
	"get", "set", "load", "parse", "save",
	"keys", "figures", "figure", "dancer", "movie", "note",
	"sync", "update", "delete", "checkinsertion", "insert",
	"applyto", "dump", (char *) 0};
enum {get, set, load, parse, save,
	  keys, figures, figur, dancer, movie, note,
	  sync, update, del, check, insert,
	  applyto, dump} index;

    if (objc < 2) {
        Tcl_WrongNumArgs(interp, 1, objv, "option ?arg ...?");
        return TCL_ERROR;
    }

    if (Tcl_GetIndexFromObj(interp, objv[1], optionsPtr, "option", 0, (int *)&index) != TCL_OK) {
        return TCL_ERROR;
    }

	switch (index) {
		case set:		return routine->Set(interp, objc, objv);
		case get:		return routine->Get(interp, objc, objv);
		case parse:
		case load:		return routine->Load(interp, objc, objv);
		case save:		return routine->Save(interp, objc, objv);

		case keys:		return routine->GetKeys(interp, objc, objv);
		case figures:	return routine->GetFigures(interp, objc, objv);
		case figur:		return routine->GetFigureInfo(interp, objc, objv);
		case note:		return routine->HandleNote(interp, objc, objv);
		case dancer:	return routine->HandleDancer(interp, objc, objv);
		case movie:		return routine->HandleMovie(interp, objc, objv);

		case sync:		return routine->Sync(interp, objc, objv);
		case update:	return routine->Update(interp, objc, objv);

		case del:		return routine->Delete(interp, objc, objv);
		case check:		return routine->CheckInsertion(interp, objc, objv);
		case insert:	return routine->Insert(interp, objc, objv);
		case applyto:	return routine->ApplyTo(interp, objc, objv);

#ifdef DEBUG
		case dump:		routine->Dump(); break;
#else
		case dump:		break;
#endif
	}

	return TCL_OK;
}

static void RoutineDelete_Cmd(ClientData clientData)
{
class Routine *routine = (Routine *) clientData;

#ifndef WIN32
	delete routine;
#endif
}

//================================================================================================

int RoutineCreate_Cmd(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
static int counter = 0;
char cmd[10];
class Routine *r;

    if (argc != 1) {
        Tcl_SetResult(interp, "wrong # of args : routine", TCL_STATIC);
        return TCL_ERROR;
    }

	r = new Routine(counter);
	sprintf(cmd, "routine%d", counter++);
	Tcl_CreateObjCommand(interp, cmd, Routine_ObjCmd, 
						 (ClientData) r, (Tcl_CmdDeleteProc *) RoutineDelete_Cmd);
	
	Tcl_SetResult(interp, cmd, TCL_VOLATILE);
	return TCL_OK;
}
